from __future__ import annotations

import binascii
import json
import typing

from ..signed_object import AbstractSignedObject
from ...types import PrivateKey, Token


class Metadata(AbstractSignedObject):
    """
    A JSON dictionary and a pointer to a Token.

    Metadata does not and should not contain an index.
    Metadata does not and should not contain a reference to the public key (but is signed by one).
    """

    def __init__(self,
                 token_pointer: bytes,
                 serialized_json_dict: bytes,
                 private_key: typing.Optional[PrivateKey] = None,
                 signature: typing.Optional[bytes] = None):
        """
        Create a new Metadata object, always specify the token it belongs to and its contents as a JSON dictionary.

        For Metadata belonging to yourself, give the token pointer, the JSON dictionary and private key and the content
        hash and signature will be automatically created.
        For Metadata of someone else or your own reloaded Metadata, give the token pointer, the JSON dictionary and
        the signature.

        :param token_pointer: a string pointing to the hash of the Token.
        :param serialized_json_dict: the JSON dictionary of metadata to tokenize.
        :param private_key: the private key of the current user, creating this new Token.
        :param signature: the signature for this Token.
        """
        self.token_pointer = token_pointer
        self.serialized_json_dict = serialized_json_dict
        super(Metadata, self).__init__(private_key, signature)

    def get_plaintext(self) -> bytes:
        """
        Serialized Metadata consists of a token pointer and serialized JSON.
        """
        return self.token_pointer + self.serialized_json_dict

    @classmethod
    def unserialize(cls, data, public_key, offset=0) -> Metadata:
        if offset != 0:
            raise RuntimeError("Offset is not supported for Metadata!")
        sig_len = public_key.get_signature_length()
        return Metadata(data[:32], data[32:-sig_len], signature=data[-sig_len:])

    @classmethod
    def create(cls, token: Token, json_dict: dict, private_key: PrivateKey) -> Metadata:
        return Metadata(token.get_hash(), json.dumps(json_dict).encode(), private_key=private_key)

    def to_database_tuple(self) -> typing.Tuple[bytes, bytes, bytes]:
        """
        Get a representation of this Metadata as three byte strings (token hash, signature and json dictionary).

        :returns: the three byte strings for database insertion.
        """
        return self.token_pointer, self.signature, self.serialized_json_dict

    @classmethod
    def from_database_tuple(cls,
                            token_pointer: bytes,
                            signature: bytes,
                            serialized_json_dict: bytes) -> Metadata:
        """
        Create a Token from a three-byte-string representation (token hash, signature and json dictionary).

        :param token_pointer: the hash of the Token.
        :param signature: the signature over the plaintext Metadata.
        :param serialized_json_dict: the serialized json dictionary of this Metadata.
        """
        return Metadata(token_pointer, serialized_json_dict, signature=signature)

    def __str__(self) -> str:
        return (f"Metadata({binascii.hexlify(self.token_pointer).decode()},\n"
                f"{self.serialized_json_dict.decode()})")



import logging
import time
from asyncio import Future

from .tunnel import CIRCUIT_STATE_CLOSING, CIRCUIT_STATE_READY
from ...requestcache import NumberCache, RandomNumberCache


class CreateRequestCache(RandomNumberCache):
    """
    Used to track outstanding create messages
    """
    def __init__(self, community, identifier, to_circuit_id, from_circuit_id, peer, to_peer):
        super().__init__(community.request_cache, "create")
        self.community = community
        self.extend_identifier = identifier
        self.to_circuit_id = to_circuit_id
        self.from_circuit_id = from_circuit_id
        self.peer = peer
        self.to_peer = to_peer

    def on_timeout(self):
        to_circuit = self.community.circuits.get(self.to_circuit_id)
        if to_circuit and to_circuit.state != CIRCUIT_STATE_READY:
            self.community.remove_relay(self.to_circuit_id)


class CreatedRequestCache(NumberCache):
    """
    Used to track outstanding created messages
    """
    def __init__(self, community, circuit_id, candidate, candidates, timeout):
        super().__init__(community.request_cache, "created", circuit_id)
        self.circuit_id = circuit_id
        self.candidate = candidate
        self.candidates = candidates
        self.timeout = timeout

    @property
    def timeout_delay(self):
        return float(self.timeout)


class RetryRequestCache(RandomNumberCache):
    """
    Used to track adding additional hops to the circuit.
    """
    def __init__(self, community, circuit, candidates, max_tries, retry_func, timeout):
        super().__init__(community.request_cache, "retry")
        self.community = community
        self.circuit = circuit
        self.candidates = candidates
        self.max_tries = max_tries
        self.retry_func = retry_func
        self.timeout = timeout
        self.logger = logging.getLogger(__name__)

    @property
    def timeout_delay(self):
        return float(self.timeout)

    def on_timeout(self):
        if self.circuit.state == CIRCUIT_STATE_CLOSING:
            return
        if not self.candidates or self.max_tries < 1:
            reason = f'timeout, {self.max_tries} tries left'
            self.community.remove_circuit(self.circuit.circuit_id, reason)
            return

        async def retry_later():
            try:
                self.retry_func(self.circuit, self.candidates, self.max_tries)
            except Exception as e:
                self.logger.info("Error encountered during 'on_timeout' (error: %s)", e)
        self.community.request_cache.register_anonymous_task("retry-later", retry_later, delay=0)


class PingRequestCache(RandomNumberCache):

    def __init__(self, community):
        super().__init__(community.request_cache, "ping")


class IPRequestCache(RandomNumberCache):

    def __init__(self, community, circuit):
        super().__init__(community.request_cache, "establish-intro")
        self.logger = logging.getLogger(__name__)
        self.circuit = circuit
        self.community = community

    def on_timeout(self):
        self.logger.info("IPRequestCache: no response on establish-intro (circuit %d)", self.circuit.circuit_id)
        self.community.remove_circuit(self.circuit.circuit_id, 'establish-intro timeout')


class RPRequestCache(RandomNumberCache):

    def __init__(self, community, rp):
        super().__init__(community.request_cache, "establish-rendezvous")
        self.logger = logging.getLogger(__name__)
        self.community = community
        self.rp = rp

    def on_timeout(self):
        self.logger.info("RPRequestCache: no response on establish-rendezvous (circuit %d)",
                         self.rp.circuit.circuit_id)
        self.rp.ready.set_result(None)
        self.community.remove_circuit(self.rp.circuit.circuit_id, 'establish-rendezvous timeout')


class PeersRequestCache(RandomNumberCache):

    def __init__(self, community, circuit, info_hash, target):
        super().__init__(community.request_cache, "peers-request")
        self.community = community
        self.circuit = circuit
        self.info_hash = info_hash
        self.target = target
        self.future = Future()
        self.register_future(self.future, RuntimeError("Peers request timeout"))

    def on_timeout(self):
        swarm = self.community.swarms.get(self.info_hash)
        if swarm and self.target:
            # This introduction point did not respond in time, so drop it.
            swarm.remove_intro_point(self.target)


class E2ERequestCache(RandomNumberCache):

    def __init__(self, community, info_hash, hop, intro_point):
        super().__init__(community.request_cache, "e2e-request")
        self.info_hash = info_hash
        self.hop = hop
        self.intro_point = intro_point


class LinkRequestCache(RandomNumberCache):

    def __init__(self, community, circuit, info_hash, hs_session_keys):
        super().__init__(community.request_cache, "link-request")
        self.circuit = circuit
        self.info_hash = info_hash
        self.hs_session_keys = hs_session_keys


class TestRequestCache(RandomNumberCache):

    def __init__(self, community, circuit):
        super().__init__(community.request_cache, "test-request")
        self.circuit = circuit
        self.ts = time.time()
        self.future = Future()
        self.register_future(self.future)



"""
Copyright (c) 2016, Maarten Everts
All rights reserved.

This source code has been ported from https://github.com/privacybydesign/gabi
The authors of this file are not -in any way- affiliated with the original authors or organizations.
"""
import unittest

from ....base import TestBase
from .....attestation.wallet.irmaexact.gabi.keys import DefaultSystemParameters, GenerateKeyPair, SignMessageBlock


class TestKeys(TestBase):

    @unittest.SkipTest  # Too slow and unused.
    def test_generate_and_sign(self):
        """
        Generate a new key and sign a message, see if the signature verifies.
        """
        privkey, pubkey = GenerateKeyPair(DefaultSystemParameters[1024], 1, 0, 0)
        sig = SignMessageBlock(privkey, pubkey, [1])

        self.assertTrue(sig.Verify(pubkey, [1]))
        self.assertTrue(sig.Randomize(pubkey).Verify(pubkey, [1]))



from asyncio import Future, TimeoutError as AsyncTimeoutError, get_event_loop, sleep, wait_for
from binascii import unhexlify
from os import environ
from random import randint
from socket import gethostbyname

# Check if we are running from the root directory
# If not, modify our path so that we can import IPv8
try:
    import ipv8
    del ipv8
except ImportError:
    import __scriptpath__  # noqa: F401

from ipv8.community import Community
from ipv8.configuration import DISPERSY_BOOTSTRAPPER, get_default_configuration
from ipv8.requestcache import NumberCache, RequestCache

from ipv8_service import IPv8


community_id = unhexlify(environ.get('INTRODUCTION_CID', '7e313685c1912a141279f8248fc8db5899c5df5a'))
count = int(environ.get('INTRODUCTION_COUNT', 100))
delay = int(environ.get('INTRODUCTION_DELAY', 10))


class MyCommunity(Community):
    community_id = community_id

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.request_cache = RequestCache()

    def send_intro_request(self, target_addr):
        packet = self.create_introduction_request(target_addr, new_style=self.network.is_new_style(target_addr))
        cache = NumberCache(self.request_cache, 'intro-req', self.global_time)
        cache.future = Future()
        self.request_cache.add(cache)
        self.endpoint.send(target_addr, packet)
        return cache.future

    def introduction_response_callback(self, peer, dist, payload):
        if self.request_cache.has('intro-req', payload.identifier):
            cache = self.request_cache.pop('intro-req', payload.identifier)
            if not cache.future.done():
                cache.future.set_result(payload)


async def main():
    configuration = get_default_configuration()
    configuration['keys'] = [{
        'alias': "my peer",
        'generation': "curve25519",
        'file': "ec.pem"
    }]
    configuration['port'] = 12000 + randint(0, 10000)
    configuration['overlays'] = []
    ipv8 = IPv8(configuration)
    await ipv8.start()

    overlay = MyCommunity(ipv8.keys['my peer'], ipv8.endpoint, ipv8.network)
    introductions = {}
    ips = {}
    for _ in range(count):
        for dns_addr in DISPERSY_BOOTSTRAPPER['init']['dns_addresses']:
            if 'tribler.org' not in dns_addr[0] and 'ip-v8.org' not in dns_addr[0]:
                continue

            try:
                ip = ips.get(dns_addr[0], gethostbyname(dns_addr[0]))
            except OSError:
                continue

            try:
                response = await wait_for(overlay.send_intro_request((ip, dns_addr[1])), timeout=5)
            except AsyncTimeoutError:
                continue

            reachable = False
            try:
                if response.wan_introduction_address != ('0.0.0.0', 0):
                    # Wait some time for puncture to be sent
                    await sleep(.1)
                    await wait_for(overlay.send_intro_request(response.wan_introduction_address), timeout=5)
                    reachable = True
            except AsyncTimeoutError:
                pass

            introductions[dns_addr] = introductions.get(dns_addr, [])
            introductions[dns_addr].append([response.wan_introduction_address,
                                            response.lan_introduction_address,
                                            reachable])

        await sleep(delay)

    with open('bootstrap_introductions.txt', 'w') as f:
        f.write('Address Peers Type')
        for dns_addr, responses in introductions.items():
            f.write(f"\n{dns_addr[0]}:{dns_addr[1]} {len([wan for wan, _, _ in responses if wan != ('0.0.0.0', 0)])} 0")
            f.write(f"\n{dns_addr[0]}:{dns_addr[1]} {len({wan for wan, _, _ in responses if wan != ('0.0.0.0', 0)})} 1")
            f.write(f"\n{dns_addr[0]}:{dns_addr[1]} {len({lan for _, lan, _ in responses if lan != ('0.0.0.0', 0)})} 3")
            f.write(f"\n{dns_addr[0]}:{dns_addr[1]} {len({wan for wan, _, reachable in responses if reachable})} 2")


if __name__ == "__main__":
    loop = get_event_loop()
    loop.run_until_complete(main())



import os
import time
from asyncio import ensure_future, get_event_loop
from random import randint

# Check if we are running from the root directory
# If not, modify our path so that we can import IPv8
try:
    import ipv8
    del ipv8
except ImportError:
    import __scriptpath__  # noqa: F401

from ipv8.community import Community
from ipv8.configuration import DISPERSY_BOOTSTRAPPER, get_default_configuration

from ipv8_service import IPv8, _COMMUNITIES


START_TIME = time.time()
LOW_EDGE = 0
LOW_EDGE_PEER = None
INSTANCES = []


class MyCommunity(Community):
    community_id = os.urandom(20)

    def started(self):
        async def check_peers():
            global INSTANCES, LOW_EDGE, LOW_EDGE_PEER, START_TIME
            if self.get_peers():
                if LOW_EDGE and self.my_peer != LOW_EDGE_PEER:
                    print("%.4f,%.4f" % (LOW_EDGE, time.time() - START_TIME))

                    async def shutdown():
                        for instance in INSTANCES:
                            await instance.stop(False)
                        get_event_loop().stop()
                    ensure_future(shutdown())
                else:
                    LOW_EDGE = time.time() - START_TIME
                    LOW_EDGE_PEER = self.my_peer
        self.register_task("check_peers", check_peers, interval=0.1, delay=0)

    def create_introduction_response(self, lan_socket_address, socket_address, identifier, introduction=None,
                                     extra_bytes=b'', prefix=None, new_style=False):
        return super(MyCommunity, self).create_introduction_response(lan_socket_address, socket_address,
                                                                     identifier, introduction, b'1', prefix, new_style)

    def create_introduction_request(self, socket_address, extra_bytes=b'', new_style=False):
        return super().create_introduction_request(socket_address, b'2', new_style)


_COMMUNITIES['MyCommunity'] = MyCommunity


async def start_communities():
    for i in [1, 2]:
        configuration = get_default_configuration()
        configuration['keys'] = [{
            'alias': "my peer",
            'generation': u"medium",
            'file': u"ec%d.pem" % i
        }]
        configuration['port'] = 12000 + randint(0, 10000)
        configuration['overlays'] = [{
            'class': 'MyCommunity',
            'key': "my peer",
            'walkers': [{
                'strategy': "RandomWalk",
                'peers': 10,
                'init': {
                    'timeout': 3.0
                }
            }],
            'bootstrappers': [DISPERSY_BOOTSTRAPPER],
            'initialize': {},
            'on_start': [('started', )]
        }]
        ipv8 = IPv8(configuration)
        await ipv8.start()
        INSTANCES.append(ipv8)


ensure_future(start_communities())
get_event_loop().run_forever()



import collections
import logging
import time
from asyncio import all_tasks, current_task, get_event_loop

from aiohttp import web

from aiohttp_apispec import docs, json_schema

from marshmallow.fields import Boolean, Float, Integer, String

from .base_endpoint import BaseEndpoint, HTTP_BAD_REQUEST, HTTP_NOT_FOUND, Response
from .schema import DefaultResponseSchema, schema
from ..peerdiscovery.discovery import DiscoveryStrategy


class DequeLogHandler(logging.Handler):

    def __init__(self, maxlen=50):
        super(DequeLogHandler, self).__init__()
        self.deque = collections.deque(maxlen=maxlen)

    def emit(self, record):
        self.deque.append(self.format(record))


class DriftMeasurementStrategy(DiscoveryStrategy):

    def __init__(self, core_update_rate):
        super(DriftMeasurementStrategy, self).__init__(None)
        self.last_measurement = time.time()
        self.history = collections.deque(maxlen=100)
        self.core_update_rate = core_update_rate
        self.overlay = type('FakeOverlay', (object,), {'get_peers': lambda: []})

    def take_step(self):
        with self.walk_lock:
            this_time = time.time()
            self.history.append((this_time, max(0.0, this_time - self.last_measurement - self.core_update_rate)))
            self.last_measurement = this_time


class AsyncioEndpoint(BaseEndpoint):
    """
    This endpoint helps with the monitoring of the Asyncio thread.
    """

    def __init__(self):
        super(AsyncioEndpoint, self).__init__()
        self.strategy = None
        self.asyncio_log_handler = None

    def setup_routes(self):
        self.app.add_routes([web.get('/drift', self.retrieve_drift),
                             web.put('/drift', self.enable_measurements),
                             web.get('/tasks', self.get_asyncio_tasks),
                             web.put('/debug', self.set_asyncio_debug),
                             web.get('/debug', self.get_asyncio_debug)])

    def enable(self):
        if not self.session:
            return False
        if not self.strategy:
            self.strategy = DriftMeasurementStrategy(self.session.walk_interval)
            with self.session.overlay_lock:
                self.session.strategies.append((self.strategy, -1))
            return True
        return True

    def disable(self):
        if not self.session:
            return False
        if self.strategy:
            with self.session.overlay_lock:
                self.session.strategies = [s for s in self.session.strategies if s[0] != self.strategy]
            self.strategy = None
            return True
        return True

    @docs(
        tags=["Asyncio"],
        summary="Measure the core drift.",
        responses={
            200: {
                "schema": schema(DriftResponse={"measurements": [
                    schema(Measurement={
                        "timestamp": Float,
                        "drift": Float
                    })
                ]})
            },
            400: {
                "schema": DefaultResponseSchema,
                "example": {"success": False, "error": "Core drift disabled."}
            }
        }
    )
    async def retrieve_drift(self, _):
        if self.strategy is None:
            return Response({"success": False, "error": "Core drift disabled."}, status=HTTP_NOT_FOUND)
        with self.strategy.walk_lock:
            return Response({"measurements": [{"timestamp": e[0], "drift": e[1]} for e in self.strategy.history]})

    @docs(
        tags=["Asyncio"],
        summary="Enable or disable drift measurements.",
        responses={
            200: {
                "schema": DefaultResponseSchema,
                "example": {"success": True}
            },
            400: {
                "schema": DefaultResponseSchema,
                "examples": {'Enable value not specified': {"success": False, "error": "incorrect parameters"}}
            }
        }
    )
    @json_schema(schema(EnableDriftRequest={
        'enable*': (Boolean, 'Whether to enable or disable measuring.'),
    }))
    async def enable_measurements(self, request):
        parameters = await request.json()
        if 'enable' not in parameters:
            return Response({"error": "incorrect parameters"}, status=HTTP_BAD_REQUEST)

        status = self.enable() if parameters.get('enable') else self.disable()
        if status:
            return Response({"success": True})
        else:
            return Response({"success": False, "error": "Session not initialized."})

    @docs(
        tags=["Asyncio"],
        summary="Return all currently running Asyncio tasks.",
        responses={
            200: {
                "schema": schema(AsyncioTasksResponse={"tasks": [
                    schema(AsyncioTask={
                        "name": String,
                        "running": Boolean,
                        "stack": [String],
                        "taskmanager": String,
                        "start_time": Float,
                        "interval": Integer
                    })
                ]})
            }
        }
    )
    async def get_asyncio_tasks(self, _):
        current = current_task()
        tasks = []
        for task in all_tasks():
            # Only in Python 3.8+ will we have a get_name function
            name = task.get_name() if hasattr(task, 'get_name') else getattr(task, 'name', f'Task-{id(task)}')

            task_dict = {"name": name,
                         "running": task == current,
                         "stack": [str(f) for f in task.get_stack()]}

            # Add info specific to tasks owner by TaskManager
            if hasattr(task, "start_time"):
                # Only TaskManager tasks have a start_time attribute
                cls, tsk = name.split(":")
                task_dict.update({"name": tsk, "taskmanager": cls, "start_time": task.start_time})
                if task.interval:
                    task_dict["interval"] = task.interval
            tasks.append(task_dict)
        return Response({"tasks": tasks})

    @docs(
        tags=["Asyncio"],
        summary="Set asyncio debug options.",
        responses={
            200: {
                "schema": DefaultResponseSchema,
                "example": {"success": True}
            },
            400: {
                "schema": DefaultResponseSchema,
                "example": {"success": False, "error": "incorrect parameters"}
            }
        }
    )
    @json_schema(schema(AsyncioDebugRequest={
        'enable': (Boolean, 'Whether to enable or disable asyncio debug mode.'),
        'slow_callback_duration': (Integer, 'Time threshold above which tasks will be logged.'),
    }))
    async def set_asyncio_debug(self, request):
        parameters = await request.json()
        if 'enable' not in parameters and 'slow_callback_duration' not in parameters:
            return Response({"success": False, "error": "incorrect parameters"}, status=HTTP_BAD_REQUEST)

        loop = get_event_loop()
        loop.slow_callback_duration = parameters.get('slow_callback_duration', loop.slow_callback_duration)

        if 'enable' in parameters:
            enable = bool(parameters['enable'])
            loop.set_debug(enable)

            # Add/remove asyncio log handler
            if enable and not self.asyncio_log_handler:
                self.asyncio_log_handler = DequeLogHandler()
                logging.getLogger('asyncio').addHandler(self.asyncio_log_handler)
            if not enable and self.asyncio_log_handler:
                logging.getLogger('asyncio').removeHandler(self.asyncio_log_handler)
                self.asyncio_log_handler = None

        return Response({"success": True})

    @docs(
        tags=["Asyncio"],
        summary="Return Asyncio log messages.",
        responses={
            200: {
                "schema": schema(AsyncioLogResponse={"debug": [
                    schema(AsyncioLogMessage={
                        "message": String
                    })
                ]})
            },
            400: {
                "schema": DefaultResponseSchema,
                "example": {"success": False, "error": "debug mode is disabled"}
            }
        }
    )
    async def get_asyncio_debug(self, _):
        loop = get_event_loop()
        messages = self.asyncio_log_handler.deque if self.asyncio_log_handler else []
        return Response({'messages': [{'message': r} for r in messages],
                         'enable': loop.get_debug(),
                         'slow_callback_duration': loop.slow_callback_duration})



from ...messaging.lazy_payload import VariablePayload, vp_compile


@vp_compile
class DiclosePayload(VariablePayload):
    msg_id = 1
    format_list = ['varlenH', 'varlenH', 'varlenH', 'varlenH']
    names = ['metadata', 'tokens', 'attestations', 'authorities']

    metadata: bytes
    tokens: bytes
    attestations: bytes
    authorities: bytes


@vp_compile
class AttestPayload(VariablePayload):
    msg_id = 2
    format_list = ['varlenH']
    names = ['attestation']

    attestation: bytes


@vp_compile
class RequestMissingPayload(VariablePayload):
    msg_id = 3
    format_list = ['I']
    names = ['known']

    known: int


@vp_compile
class MissingResponsePayload(VariablePayload):
    msg_id = 4
    format_list = ['raw']
    names = ['tokens']

    tokens: bytes



import logging
from asyncio import CancelledError, gather
from contextlib import contextmanager, suppress
from random import random
from threading import Lock
from typing import Generator, Iterable, Optional, Type

from .taskmanager import TaskManager


class NumberCache(object):

    def __init__(self, request_cache, prefix, number):
        assert isinstance(number, int), type(number)
        assert isinstance(prefix, str), type(prefix)

        super(NumberCache, self).__init__()
        self._logger = logging.getLogger(self.__class__.__name__)

        if request_cache.has(prefix, number):
            raise RuntimeError("This number is already in use '%s'" % number)

        self._prefix = prefix
        self._number = number

        self._managed_futures = []

    def register_future(self, future, on_timeout=None):
        """
        Register a future for this Cache that will be canceled when this Cache times out.

        :param future: the future to register for this instance
        :type future: Future
        :param on_timeout: The value to which the future is to be set when a timeout occurs. If the value is an
                           instance of Exception, future.set_exception will be called instead of future.set_result
        :type on_timeout: Object
        :returns: None
        """
        self._managed_futures.append((future, on_timeout))

    @property
    def managed_futures(self):
        return self._managed_futures

    @property
    def prefix(self):
        return self._prefix

    @property
    def number(self):
        return self._number

    @property
    def timeout_delay(self):
        return 10.0

    def on_timeout(self):
        pass

    def __str__(self):
        return "<%s %s-%d>" % (self.__class__.__name__, self.prefix, self.number)


class RandomNumberCache(NumberCache):

    def __init__(self, request_cache, prefix):
        assert isinstance(prefix, str), type(prefix)

        # find an unclaimed identifier
        number = RandomNumberCache.find_unclaimed_identifier(request_cache, prefix)
        super(RandomNumberCache, self).__init__(request_cache, prefix, number)

    @classmethod
    def find_unclaimed_identifier(cls, request_cache, prefix):
        for _ in range(1000):
            number = int(random() * 2 ** 16)
            if not request_cache.has(prefix, number):
                break
        else:
            raise RuntimeError("Could not find a number that isn't in use")

        return number


class RequestCache(TaskManager):

    def __init__(self):
        """
        Creates a new RequestCache instance.
        """
        super(RequestCache, self).__init__()

        self._logger = logging.getLogger(self.__class__.__name__)

        self._identifiers = dict()
        self.lock = Lock()
        self._shutdown = False

        self._timeout_override: Optional[float] = None
        """
        If not None, this specifies the timeout to use instead of the one defined in
        the ``timeout_delay`` of a ``NumberCache``.
        This is used internally for ``passthrough()``, don't modify this directly!
        """

        self._timeout_filters: Optional[Iterable[Type[NumberCache]]] = None
        """
        If not None, this specifies the ``NumberCache`` (sub)classes to apply the
        timeout override for.
        This is used internally for ``passthrough()``, don't modify this directly!
        """

    def add(self, cache):
        """
        Add CACHE into this RequestCache instance.

        Returns CACHE when CACHE.identifier was not yet added, otherwise returns None.
        """
        assert isinstance(cache, NumberCache), type(cache)
        assert isinstance(cache.number, int), type(cache.number)
        assert isinstance(cache.prefix, str), type(cache.prefix)
        assert isinstance(cache.timeout_delay, (int, float)), type(cache.timeout_delay)
        assert cache.timeout_delay > 0.0, cache.timeout_delay

        with self.lock:
            if self._shutdown:
                self._logger.warning("Dropping %s due to shutdown!", str(cache))
                for f, _ in cache.managed_futures:
                    f.cancel()
                return None

            identifier = self._create_identifier(cache.number, cache.prefix)
            if identifier in self._identifiers:
                self._logger.error("add with duplicate identifier \"%s\"", identifier)
                return None

            else:
                self._logger.debug("add %s", cache)
                self._identifiers[identifier] = cache

                timeout_delay = cache.timeout_delay
                if self._timeout_override is not None:
                    # Only overwrite the timeout if an overwrite is set.
                    if (self._timeout_filters is None
                            or any(issubclass(cache.__class__, f) for f in self._timeout_filters)):
                        # Only overwrite the timeout if no class filters are set.
                        # Otherwise, only overwrite the timeout if the cache class is in the filter.
                        timeout_delay = self._timeout_override

                self.register_task(cache, self._on_timeout, cache, delay=timeout_delay)
                return cache

    def has(self, prefix, number):
        """
        Returns True when IDENTIFIER is part of this RequestCache.
        """
        assert isinstance(number, int), type(number)
        assert isinstance(prefix, str), type(prefix)
        return self._create_identifier(number, prefix) in self._identifiers

    def get(self, prefix, number):
        """
        Returns the Cache associated with IDENTIFIER when it exists, otherwise returns None.
        """
        assert isinstance(number, int), type(number)
        assert isinstance(prefix, str), type(prefix)
        return self._identifiers.get(self._create_identifier(number, prefix))

    def pop(self, prefix, number):
        """
        Returns the Cache associated with IDENTIFIER, and removes it from this RequestCache, when it exists, otherwise
        raises a KeyError exception.
        """
        assert isinstance(number, int), type(number)
        assert isinstance(prefix, str), type(prefix)

        identifier = self._create_identifier(number, prefix)
        cache = self._identifiers.pop(identifier)
        self.cancel_pending_task(cache)
        return cache

    @contextmanager
    def passthrough(self,  # pylint: disable=W1113
                    cls_filter: Optional[Type[NumberCache]] = None, *filters: Type[NumberCache],
                    timeout: float = 0.0) -> Generator:
        """
        A contextmanager that overwrites the timeout_delay of added NumberCaches in its scope.
        This can be used to shorten or eliminate timeouts of external code.

        ---
        Example 1: Eliminating timeouts, regardless of ``cache.timeout_delay``
        ---

         .. code-block :: Python

            with request_cache.passthrough():
                request_cache.add(cache)  # This will instantly timeout (once the main thread is yielded).

            with request_cache.passthrough():
                # Any internal call to request_cache.add() will also be instantly timed out.
                await some_function_that_uses_request_cache()

        ---
        Example 2: Modifying timeouts, regardless of ``cache.timeout_delay``
        ---

         .. code-block :: Python

            with request_cache.passthrough(timeout=0.1):
                request_cache.add(cache)  # This will timeout after 0.1 seconds.

        ---
        Example 3: Filtering for specific classes
        ---

         .. code-block :: Python

            # Only MyCacheClass, MyOtherCacheClass, YetAnotherCacheClass will have their timeout changed to 4 seconds.
            with request_cache.passthrough(MyCacheClass, MyOtherCacheClass, YetAnotherCacheClass, timeout=4.0):
                request_cache.add(cache)

        :param cls_filter: An optional class filter to specify which classes the timeout override needs to apply to.
        :param filters: Additional class filters to specify which classes the timeout override needs to apply to.
        :param timeout: The timeout in seconds to use for the ``NumberCache`` instances this applies to.
        :returns: A context manager (compatible with ``with``).
        """
        self._timeout_override = timeout
        self._timeout_filters = None if cls_filter is None else [cls_filter] + list(filters)
        try:
            yield
        finally:
            self._timeout_override = None
            self._timeout_filters = None

    def _on_timeout(self, cache):
        """
        Called CACHE.timeout_delay seconds after CACHE was added to this RequestCache.

        _on_timeout is called for every Cache, except when it has been popped before the timeout expires.  When called
        _on_timeout will CACHE.on_timeout().
        """
        assert isinstance(cache, NumberCache), type(cache)

        self._logger.debug("timeout on %s", cache)

        # the on_timeout call could have already removed the identifier from the cache using pop
        identifier = self._create_identifier(cache.number, cache.prefix)
        if identifier in self._identifiers:
            self._identifiers.pop(identifier)

        cache.on_timeout()

        for future, on_timeout in cache.managed_futures:
            if not future.done():
                if isinstance(on_timeout, Exception):
                    future.set_exception(on_timeout)
                else:
                    future.set_result(on_timeout)

        self.cancel_pending_task(cache)

    def _create_identifier(self, number, prefix):
        return u"%s:%d" % (prefix, number)

    def clear(self):
        """
        Clear the cache, canceling all pending tasks.

        """
        self._logger.debug("Clearing %s [%s]", self, len(self._identifiers))
        tasks = self.cancel_all_pending_tasks()
        self._identifiers.clear()
        return tasks

    async def shutdown(self):
        """
        Clear the cache, cancel all pending tasks and disallow new caches being added.
        """
        with self.lock:
            # Don't call TaskManager.shutdown_task_manager here since
            # we don't want to await stuff while holding a non-async lock.
            with self._task_lock:
                self._shutdown = True
                tasks = self.cancel_all_pending_tasks()

            for cache in self._identifiers.values():
                # Cancel all managed futures, and suppress the CancelledErrors
                for future, _ in cache.managed_futures:
                    future.cancel()
            self._identifiers.clear()

        if tasks:
            with suppress(CancelledError):
                await gather(*tasks)



from base64 import b64encode

from aiohttp import web

from aiohttp_apispec import docs

from marshmallow.fields import Integer, List, String

from .base_endpoint import BaseEndpoint, Response
from .schema import schema


class NetworkEndpoint(BaseEndpoint):
    """
    This endpoint is responsible for handing all requests regarding the state of the network.
    """

    def setup_routes(self):
        self.app.add_routes([web.get('', self.retrieve_peers)])

    @docs(
        tags=["Network"],
        summary="Return a list of all known peers.",
        responses={
            200: {
                "schema": schema(PeersResponse={
                    "peers": [schema(Peer={
                        "ip": String,
                        "port": Integer,
                        "public_key": String,
                        "services": List(String),
                    })]
                })
            }
        }
    )
    async def retrieve_peers(self, _):
        network = self.session.network
        peer_list = network.verified_peers
        return Response({"peers": {
            b64encode(peer.mid).decode('utf-8'): {
                "ip": peer.address[0],
                "port": peer.address[1],
                "public_key": b64encode(peer.public_key.key_to_bin()).decode('utf-8'),
                "services": [b64encode(s).decode('utf-8') for s in network.get_services_for_peer(peer)]
            }
            for peer in peer_list
        }})



from binascii import unhexlify

from ....base import TestBase
from .....attestation.wallet.pengbaorange.boudot import EL, SQR
from .....attestation.wallet.primitives.structs import BonehPrivateKey
from .....attestation.wallet.primitives.value import FP2Value


class TestBoudot(TestBase):

    pk = BonehPrivateKey.unserialize(unhexlify("0109649b2a7d7992b008d1010958f4d560346bb392330109117572cef25"
                                               "23be2fe0109264c97f1868fd4b18b010908792e2bf8bcaa077f0108baa1"
                                               "b2ff06a0cfd50104c772a78f")).public_key()
    bitspace = 32
    a, b = 0, 18
    r, ra = 2, 3

    # Commitment generated with above parameters for message `7`
    el = EL.unserialize(unhexlify("02012101c42b8cb90f9de7b77897d0df644410cf8581a7381392562164b5e0da43d3e47f0121012d725dd0b513efcfa5ba8b3f982d608a59011a25626138c43fb537a3faf168ff01210710ae32e43e779edde25f437d9110433e16069ce04e488d0c6c242fba6c50be7c012096b92ee85a89f7e7d2dd459fcc16b0452c808d12b130b66bb3adaea4de5c0fb6"))[0]

    # Commitment generated with above parameters for message `16`
    sqr = SQR.unserialize(unhexlify("0109649b2a7d7992b008d101092ef4b2b3890136f39d0109300f2cb5896dcadc4901012a0a16027d"
                                    "ee86ae81c4cfc7f529f46261608252ab4daa517d68e21720790e4199f78a1929dbd013a105b6012a"
                                    "0285809f7ba1aba0714188435457007cbe08eaaf87506267fdb8c0ad262dfb6c15ca968e206a9777"
                                    "949e012102af10fec9115c85e85cd7d7aef94a65d3cc6c7357642706e8a37bb0ecd804845b0120ab"
                                    "c43fb24457217a1735f5ebbe529974f31b1cd5d909c1ba28deec3b36011fa3"))[0]

    def test_el_serialize(self):
        """
        Check if Boudot equality checks are correctly serialized.
        """
        el = EL(68174117739048401651990398043836840253231911962332258219191049320869958258614,
                818089412868580819823884776526042083038782943547987098630292591850439499103868,
                -136348235478096803303980796087673680506463823924664516192465696382710419908863,
                204522353217145204955971194131510520759695735886996774897791958142625032430719)
        self.assertEqual(self.el, el)

    def test_el_equal(self):
        """
        Check if the Boudot commitment equality holds.
        """
        m = 7  # Message

        # Commitment
        c = self.pk.g.intpow(m) * self.pk.h.intpow(self.r)
        c1 = c // (self.pk.g.intpow(self.a - 1))
        c2 = self.pk.g.intpow(self.b + 1) // c

        # Shadow commitment
        ca = c1.intpow(self.b - m + 1) * self.pk.h.intpow(self.ra)

        # Check
        self.assertTrue(self.el.check(self.pk.g, self.pk.h, c1, self.pk.h, c2, ca))

    def test_el_modify_commitment(self):
        """
        Check if the Boudot commitment equality fails if the commitment message changes.
        """
        m = 7  # Message
        fake = 8

        # Commitment
        c = self.pk.g.intpow(fake) * self.pk.h.intpow(self.r)
        c1 = c // (self.pk.g.intpow(self.a - 1))
        c2 = self.pk.g.intpow(self.b + 1) // c

        # Shadow commitment
        ca = c1.intpow(self.b - m + 1) * self.pk.h.intpow(self.ra)

        # Check
        self.assertFalse(self.el.check(self.pk.g, self.pk.h, c1, self.pk.h, c2, ca))

    def test_el_modify_shadow_commitment(self):
        """
        Check if the Boudot commitment equality fails if the shadow commitment message changes.
        """
        m = 7  # Message
        fake = 8

        # Commitment
        c = self.pk.g.intpow(m) * self.pk.h.intpow(self.r)
        c1 = c // (self.pk.g.intpow(self.a - 1))
        c2 = self.pk.g.intpow(self.b + 1) // c

        # Shadow commitment
        ca = c1.intpow(self.b - fake + 1) * self.pk.h.intpow(self.ra)

        # Check
        self.assertFalse(self.el.check(self.pk.g, self.pk.h, c1, self.pk.h, c2, ca))

    def test_el_modify_commitments(self):
        """
        Check if the Boudot commitment equality fails if the shadow+commitment messages change.
        """
        fake = 8

        # Commitment
        c = self.pk.g.intpow(fake) * self.pk.h.intpow(self.r)
        c1 = c // (self.pk.g.intpow(self.a - 1))
        c2 = self.pk.g.intpow(self.b + 1) // c

        # Shadow commitment
        ca = c1.intpow(self.b - fake + 1) * self.pk.h.intpow(self.ra)

        # Check
        self.assertFalse(self.el.check(self.pk.g, self.pk.h, c1, self.pk.h, c2, ca))

    def test_sqr(self):
        """
        Check if the Boudot commitment-is-square holds.
        """
        sqr = SQR.create(4, 81, self.pk.g, self.pk.h, self.b, self.bitspace)
        self.assertTrue(sqr.check(self.pk.g, self.pk.h, self.pk.g.intpow(16) * self.pk.h.intpow(81)))

    def test_sqr_modify_commitment(self):
        """
        Check if the Boudot commitment-is-square fails if the commitment message changes.
        """
        sqr = SQR.create(9, 81, self.pk.g, self.pk.h, self.b, self.bitspace)
        self.assertFalse(sqr.check(self.pk.g, self.pk.h, self.pk.g.intpow(16) * self.pk.h.intpow(81)))

    def test_sqr_modify_shadow_commitment(self):
        """
        Check if the Boudot commitment-is-square fails if the shadow commitment message changes.
        """
        sqr = SQR.create(4, 81, self.pk.g, self.pk.h, self.b, self.bitspace)
        self.assertFalse(sqr.check(self.pk.g, self.pk.h, self.pk.g.intpow(9) * self.pk.h.intpow(81)))

    def test_sqr_serialize(self):
        """
        Check if Boudot commitment-is-square checks are correctly serialized.
        """
        el = EL(77692238748522549651683873494184958254240308467748145113388211342290056191907,
                310768954994090198606735493976739833016961233870992580453552845369160224769115,
                1378784829646587776157777333351527905785206201874845030390922605077647796646962410317344630260602014,
                -5515139318586351104624816262067481296619038413746351139945096955324703586833572052593411867046643126)
        sqr = SQR(FP2Value(self.pk.g.mod, 866182580282760557469, 886537163949459823689), el)

        self.assertEqual(self.sqr, sqr)



from hashlib import sha256, sha512

from ....util import int2byte


def to_ascii(value):
    if isinstance(value, str):
        return b''.join(int2byte(ord(c)) for c in value)
    return value


def sha512_as_int(value):
    """
    Convert a SHA512 hash to an integer.
    """
    out = 0
    hashed = sha512(to_ascii(value)).digest()
    for i in range(len(hashed)):
        out <<= 8
        out |= hashed[i]
    return out


def sha256_as_int(value):
    """
    Convert a SHA256 hash to an integer.
    """
    out = 0
    hashed = sha256(to_ascii(value)).digest()
    for i in range(len(hashed)):
        out <<= 8
        out |= hashed[i]
    return out


def sha256_4_as_int(value):
    """
    Convert a SHA256 4 byte hash to an integer.
    """
    out = 0
    hashed = sha256(to_ascii(value)).digest()[:4]
    for i in range(len(hashed)):
        out <<= 8
        out |= hashed[i]
    return out



from ....base import TestBase
from .....attestation.wallet.primitives.value import FP2Value


class TestFP2Value(TestBase):

    def test_add_unary(self):
        """
        Check if 2 + 5 = 7 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 2)
        b = FP2Value(11, 5)

        self.assertEqual(a + b, FP2Value(11, 7))

    def test_add_unary_mod(self):
        """
        Check if 5 + 7 = 1 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 5)
        b = FP2Value(11, 7)

        self.assertEqual(a + b, FP2Value(11, 1))

    def test_add_x(self):
        """
        Check if 2x + 5x = 7x mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, b=2)
        b = FP2Value(11, b=5)

        self.assertEqual(a + b, FP2Value(11, b=7))

    def test_add_x_mod(self):
        """
        Check if 5x + 7x = 1x mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, b=5)
        b = FP2Value(11, b=7)

        self.assertEqual(a + b, FP2Value(11, b=1))

    def test_add_x2_mod(self):
        """
        Check if x^2 + x^2 = 9x + 9 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, c=1)
        b = FP2Value(11, c=1)

        self.assertEqual(a + b, FP2Value(11, 9, 9))

    def test_add_combined_mod(self):
        """
        Check if x^2 + 5x + 5 + x^2 + 7x + 7 = 10x + 10 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 5, 5, 1)
        b = FP2Value(11, 7, 7, 1)

        self.assertEqual(a + b, FP2Value(11, 10, 10))

    def test_sub_unary(self):
        """
        Check if 5 - 2 = 3 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 5)
        b = FP2Value(11, 2)

        self.assertEqual(a - b, FP2Value(11, 3))

    def test_sub_unary_mod(self):
        """
        Check if 2 - 5 = 8 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 2)
        b = FP2Value(11, 5)

        self.assertEqual(a - b, FP2Value(11, 8))

    def test_sub_x(self):
        """
        Check if 5x - 2x = 3x mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, b=5)
        b = FP2Value(11, b=2)

        self.assertEqual(a - b, FP2Value(11, b=3))

    def test_sub_x_mod(self):
        """
        Check if 2x + 5x = 8x mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, b=2)
        b = FP2Value(11, b=5)

        self.assertEqual(a - b, FP2Value(11, b=8))

    def test_sub_x2_mod(self):
        """
        Check if x^2 - (x^2 + 1) = 10 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, c=1)
        b = FP2Value(11, 1, 0, 1)

        self.assertEqual(a - b, FP2Value(11, 10))

    def test_sub_combined_mod(self):
        """
        Check if x^2 + 5x + 5 - (x^2 + 7x + 7) = 9x + 9 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 5, 5, 1)
        b = FP2Value(11, 7, 7, 1)

        self.assertEqual(a - b, FP2Value(11, 9, 9))

    def test_mul_unary(self):
        """
        Check if 2 * 5 = 10 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 2)
        b = FP2Value(11, 5)

        self.assertEqual(a * b, FP2Value(11, 10))

    def test_mul_unary_mod(self):
        """
        Check if 3 * 5 = 4 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 3)
        b = FP2Value(11, 5)

        self.assertEqual(a * b, FP2Value(11, 4))

    def test_mul_x(self):
        """
        Check if 2x * 5x = 10x^2 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, b=2)
        b = FP2Value(11, b=5)

        self.assertEqual(a * b, FP2Value(11, c=10))

    def test_mul_x_mod(self):
        """
        Check if 3x * 5x = 4x^2 = 7x + 7 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, b=3)
        b = FP2Value(11, b=5)

        self.assertEqual(a * b, FP2Value(11, c=4))
        self.assertEqual(a * b, FP2Value(11, 7, 7))

    def test_mul_combined_mod(self):
        """
        Check if (5) * (x^2 + x + 1) = 5x^2 + 5x + 5 = 0 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 5)
        b = FP2Value(11, 1, 1, 1)

        self.assertEqual(a * b, FP2Value(11, 5, 5, 5))
        self.assertEqual(a * b, FP2Value(11, 0))

    def test_div_unary(self):
        """
        Check if 9 / 3 = 3 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 9)
        b = FP2Value(11, 3)

        self.assertEqual(a // b, FP2Value(11, 3))

    def test_div_unary_mod(self):
        """
        Check if 4 / 3  = 5 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 4)
        b = FP2Value(11, 3)

        self.assertEqual((a // b).normalize(), FP2Value(11, 5))

    def test_div_x(self):
        """
        Check if 9x / 3x = 3 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, b=9)
        b = FP2Value(11, b=3)

        self.assertEqual((a // b).normalize(), FP2Value(11, 3))

    def test_div_x_mod(self):
        """
        Check if 4x / 3x = 5 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, b=4)
        b = FP2Value(11, b=3)

        self.assertEqual((a // b).normalize(), FP2Value(11, 5))

    def test_div_combined_mod(self):
        """
        Check if (x^2 + 2x + 1) / (x + 1) = x + 1 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 1, 2, 1)
        b = FP2Value(11, 1, 1)

        self.assertEqual((a // b).normalize(), FP2Value(11, 1, 1))

    def test_intpow_unary_p2(self):
        """
        Check if 2^2 = 4 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 2)

        self.assertEqual(a.intpow(2), FP2Value(11, 4))

    def test_intpow_unary_p3(self):
        """
        Check if 2^3 = 8 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 2)

        self.assertEqual(a.intpow(3), FP2Value(11, 8))

    def test_intpow_unary_p4_mod(self):
        """
        Check if 2^4 = 5 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 2)

        self.assertEqual(a.intpow(4), FP2Value(11, 5))

    def test_intpow_x_p2(self):
        """
        Check if x^2 = x^2 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, b=1)

        self.assertEqual(a.intpow(2), FP2Value(11, c=1))

    def test_intpow_combined_p2(self):
        """
        Check if (x + 1)^2 = x^2 + 2x + 1 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 1, 1)

        self.assertEqual(a.intpow(2), FP2Value(11, 1, 2, 1))

    def test_intpow_negative(self):
        """
        Check if (16)^-2 = 8 mod 23 mod x^2 + x + 1.
        """
        a = FP2Value(23, 16)

        self.assertEqual(a.intpow(-2), FP2Value(23, 8))

    def test_inverse(self):
        """
        Check if (4/3)^-1 = 3/4 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 4, aC=3)

        self.assertEqual(a.inverse(), FP2Value(11, 3, aC=4))

    def test_inverse_normalized(self):
        """
        Check if (4/3)^-1 = 9 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 4, aC=3)

        self.assertEqual(a.inverse().normalize(), FP2Value(11, 9))

    def test_wp_compress_simple(self):
        """
        Check if (2 + 4x)/(1 + 2x) = 2 mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 2, 4, aC=1, bC=2)

        self.assertEqual(a.wp_compress(), FP2Value(11, 2))

    def test_wp_compress_complex(self):
        """
        Check if (2 + 6x)/(1 + 2x) = 7 + 8x mod 11 mod x^2 + x + 1.
        """
        a = FP2Value(11, 2, 6, aC=1, bC=2)

        self.assertEqual(a.wp_compress(), FP2Value(11, 7, 8))

    def test_str_zero(self):
        """
        Check if (0x^2 + 0x + 0)/(0x^2 + 0x + 1) is formatted as "0"
        """
        self.assertEqual(str(FP2Value(11, 0, 0, 0, 1, 0, 0)), '0')

    def test_str_nozero(self):
        """
        Check if (0x^2 + x + 0)/(0x^2 + 0x + 1) is formatted as "x"
        """
        self.assertEqual(str(FP2Value(11, 0, 1, 0, 1, 0, 0)), 'x')

    def test_str_nozero_div(self):
        """
        Check if (0x^2 + x + 0)/(0x^2 + 0x + 2) is formatted as "(x)/(2)"
        """
        self.assertEqual(str(FP2Value(11, 0, 1, 0, 2, 0, 0)), '(x)/(2)')

    def test_str_positive_coefficient(self):
        """
        Check if (0x^2 + 2x + 0)/(0x^2 + 0x + 1) is formatted as "2x"
        """
        self.assertEqual(str(FP2Value(11, 0, 2, 0, 1, 0, 0)), '2x')

    def test_eq_other(self):
        """
        Check if an FP2Value doesn't magically equal some unrelated object.
        """
        self.assertFalse(FP2Value(11, 2).__eq__(2))



from base64 import b64encode
from collections import deque
from struct import unpack
from time import time
from typing import Any, Dict, Optional, Type, Union

from .keyvault.crypto import default_eccrypto
from .keyvault.keys import Key
from .messaging.interfaces.udp.endpoint import UDPv4Address, UDPv6Address
from .types import Address


class DirtyDict(dict):
    """
    Dictionary that becomes dirty when elements are changed.
    """
    def __init__(self, **kwargs):
        super(DirtyDict, self).__init__(**kwargs)
        self.dirty = True

    def __setitem__(self, key, value):
        super(DirtyDict, self).__setitem__(key, value)
        self.dirty = True

    def update(self, mapping, **kwargs):
        super(DirtyDict, self).update(mapping, **kwargs)
        self.dirty = True

    def clear(self):
        super(DirtyDict, self).clear()
        self.dirty = True

    def pop(self, key):
        out = super(DirtyDict, self).pop(key)
        self.dirty = True
        return out

    def popitem(self):
        out = super(DirtyDict, self).popitem()
        self.dirty = True
        return out


class Peer(object):

    INTERFACE_ORDER = [UDPv6Address, UDPv4Address, tuple]

    def __init__(self, key: Union[Key, bytes], address: Optional[Address] = None, intro: bool = True) -> None:
        """
        Create a new Peer.

        :param key: the peer's Key (mostly public) or public key bin
        :param address: the address object for this peer (e.g. ("1.2.3.4", 0) for IPv4 over UDP)
        :param intro: is this peer suggested to us (otherwise it contacted us)
        """
        if not isinstance(key, Key):
            self.key: Key = default_eccrypto.key_from_public_bin(key)
        else:
            self.key = key  # type:ignore
        self.mid = self.key.key_to_hash()
        self.public_key = self.key.pub()
        self._addresses = DirtyDict()
        if address is not None:
            self._addresses[address.__class__] = address
        self._address = address
        self.creation_time = time()
        self.last_response = 0 if intro else time()
        self._lamport_timestamp = 0
        self.pings: deque = deque(maxlen=5)
        self.new_style_intro = False

    @property
    def addresses(self) -> Dict[Type[Address], Address]:
        """
        Retrieve the addresses belonging to this Peer.

        You are not allowed to set this addresses dict for a Peer manually.
        You can change the dictionary itself by setting its items or calling its functions, for example ``update()``.
        """
        return self._addresses

    @property
    def address(self) -> Address:
        """
        Retrieve the preferred address for this Peer.

        If you want to manually select the interface, use the ``.addresses`` dictionary instead.
        """
        if self._addresses.dirty:
            self._update_preferred_address()
        return self._address or UDPv4Address("0.0.0.0", 0)

    @address.setter
    def address(self, value: Address) -> None:
        """
        Alias of ``add_address(value)``.
        """
        self.add_address(value)

    def add_address(self, value: Any) -> None:
        """
        Add a known address for this Peer.

        Any object can form an address, but only one type of address can be used per object type.
        For example (normally A, B and C are ``namedtuple`` types):

         - Adding instances A(1), B(2) leads to addresses {A: A(1), B: B(2)}
         - Adding instances A(1), B(2), A(3) leads to addresses {A: A(3), B: B(2)}
        """
        self._addresses[value.__class__] = value
        self._update_preferred_address()

    def _update_preferred_address(self) -> None:
        """
        Update the current address to be the most preferred.
        """
        for interface in self.INTERFACE_ORDER:
            if interface in self._addresses:
                self._address = self._addresses[interface]
                break
        self._addresses.dirty = False

    def get_median_ping(self) -> Optional[float]:
        """
        Get the median ping time of this peer.

        :return: the median ping or None if no measurements were performed yet
        :rtype: float or None
        """
        if not self.pings:
            return None
        sorted_pings = sorted(self.pings)
        if len(sorted_pings) % 2 == 0:
            return (sorted_pings[len(sorted_pings) // 2 - 1] + sorted_pings[len(sorted_pings) // 2]) / 2
        else:
            return sorted_pings[len(sorted_pings) // 2]

    def get_average_ping(self) -> Optional[float]:
        """
        Get the average ping time of this peer.

        :return: the average ping or None if no measurements were performed yet
        :rtype: float or None
        """
        if not self.pings:
            return None
        return sum(self.pings) / len(self.pings)

    def update_clock(self, timestamp: int) -> None:
        """
        Update the Lamport timestamp for this peer. The Lamport clock dictates that the current timestamp is
        the maximum of the last known and the most recently delivered timestamp. This is useful when messages
        are delivered asynchronously.

        We also keep a real time timestamp of the last received message for timeout purposes.

        :param timestamp: a received timestamp
        """
        self._lamport_timestamp = max(self._lamport_timestamp, timestamp)
        self.last_response = time()  # This is in seconds since the epoch

    def get_lamport_timestamp(self) -> int:
        return self._lamport_timestamp

    def __hash__(self) -> int:
        as_long, = unpack(">Q", self.mid[:8])
        return as_long

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Peer):
            return False
        return self.public_key.key_to_bin() == other.public_key.key_to_bin()

    def __ne__(self, other: object) -> bool:
        if not isinstance(other, Peer):
            return True
        return self.public_key.key_to_bin() != other.public_key.key_to_bin()

    def __str__(self) -> str:
        return 'Peer<%s:%d, %s>' % (self.address[0], self.address[1], b64encode(self.mid).decode('utf-8'))



from ...base import TestBase
from ...mocking.endpoint import AutoMockEndpoint
from ....messaging.interfaces.network_stats import NetworkStat
from ....messaging.interfaces.statistics_endpoint import StatisticsEndpoint


class NoSendEndpoint(AutoMockEndpoint):

    def send(self, socket_address, packet):
        pass


class TestStatisticsEndpoint(TestBase):

    def setUp(self):
        super().setUp()
        self.raw_ep = NoSendEndpoint()
        self.stats_ep = StatisticsEndpoint(self.raw_ep)

        self.prefix = b'0' * 22
        self.msg_id = b'\x01'
        self.msg_num = self.msg_id[0]
        self.fake_addr = ("0.0.0.0", 0)

    async def test_getattribute(self):
        """
        Check if calls to the underlying Endpoint are forwarded.

        Note: ``wan_address`` is specific to ``AutoMockEndpoint``.
        """
        self.assertEqual(self.raw_ep.wan_address, self.stats_ep.wan_address)

    async def test_capture_send_enabled(self):
        """
        Check if send calls are registered when the prefix is enabled.
        """
        self.stats_ep.enable_community_statistics(self.prefix, True)
        self.stats_ep.send(self.fake_addr, self.prefix + self.msg_id + b'Hello World!')

        statistics = self.stats_ep.get_statistics(self.prefix)

        self.assertIsNotNone(statistics)
        self.assertIn(self.msg_num, statistics)
        self.assertEqual(1, statistics[self.msg_num].num_up)
        self.assertEqual(0, statistics[self.msg_num].num_down)

    async def test_no_capture_send_disabled(self):
        """
        Check if send calls are not registered when the prefix is disabled.
        """
        self.stats_ep.enable_community_statistics(self.prefix, False)
        self.stats_ep.send(self.fake_addr, self.prefix + self.msg_id + b'Hello World!')

        statistics = self.stats_ep.get_statistics(self.prefix)

        self.assertIsNotNone(statistics)
        self.assertFalse(statistics)

    async def test_capture_receive_enabled(self):
        """
        Check if receive calls are registered when the prefix is enabled.
        """
        self.stats_ep.enable_community_statistics(self.prefix, True)
        self.stats_ep.on_packet((self.fake_addr, self.prefix + self.msg_id + b'Hello World!'))

        statistics = self.stats_ep.get_statistics(self.prefix)

        self.assertIsNotNone(statistics)
        self.assertIn(self.msg_num, statistics)
        self.assertEqual(0, statistics[self.msg_num].num_up)
        self.assertEqual(1, statistics[self.msg_num].num_down)

    async def test_no_capture_receive_disabled(self):
        """
        Check if receive calls are not registered when the prefix is disabled.
        """
        self.stats_ep.enable_community_statistics(self.prefix, False)
        self.stats_ep.on_packet((self.fake_addr, self.prefix + self.msg_id + b'Hello World!'))

        statistics = self.stats_ep.get_statistics(self.prefix)

        self.assertIsNotNone(statistics)
        self.assertFalse(statistics)

    async def test_get_sent(self):
        """
        Check if sent messages are correctly logged for an unknown and known prefix.
        """
        stats = NetworkStat(self.msg_num)
        stats.num_up = 2
        stats.bytes_up = 42
        self.stats_ep.statistics = {self.prefix: {self.msg_num: stats}}
        unknown_prefix = b'some other prefix'

        self.assertEqual(0, self.stats_ep.get_message_sent(unknown_prefix))
        self.assertEqual(2, self.stats_ep.get_message_sent(self.prefix))
        self.assertEqual(0, self.stats_ep.get_bytes_sent(unknown_prefix))
        self.assertEqual(42, self.stats_ep.get_bytes_sent(self.prefix))

    async def test_get_received(self):
        """
        Check if received messages are correctly logged for an unknown and known prefix.
        """
        stats = NetworkStat(self.msg_num)
        stats.num_down = 2
        stats.bytes_down = 42
        self.stats_ep.statistics = {self.prefix: {self.msg_num: stats}}
        unknown_prefix = b'some other prefix'

        self.assertEqual(0, self.stats_ep.get_message_received(unknown_prefix))
        self.assertEqual(2, self.stats_ep.get_message_received(self.prefix))
        self.assertEqual(0, self.stats_ep.get_bytes_received(unknown_prefix))
        self.assertEqual(42, self.stats_ep.get_bytes_received(self.prefix))

    async def test_get_sent_excluded(self):
        """
        Check if excluded sent messages are not counted.
        """
        stats_intro, stats_puncture, stats_deprecated = NetworkStat(245), NetworkStat(249), NetworkStat(235)
        stats_intro.num_up = stats_puncture.num_up = stats_deprecated.num_up = 1
        self.stats_ep.statistics = {self.prefix: {245: stats_intro, 249: stats_puncture, 235: stats_deprecated}}

        self.assertEqual(0, self.stats_ep.get_message_sent(self.prefix))
        self.assertEqual(3, self.stats_ep.get_message_sent(self.prefix, True, True, True))
        self.assertEqual(1, self.stats_ep.get_message_sent(self.prefix, include_introduction=True))
        self.assertEqual(1, self.stats_ep.get_message_sent(self.prefix, include_puncture=True))
        self.assertEqual(1, self.stats_ep.get_message_sent(self.prefix, include_deprecated=True))

    async def test_get_received_excluded(self):
        """
        Check if excluded received messages are not counted.
        """
        stats_intro, stats_puncture, stats_deprecated = NetworkStat(245), NetworkStat(249), NetworkStat(235)
        stats_intro.num_down = stats_puncture.num_down = stats_deprecated.num_down = 1
        self.stats_ep.statistics = {self.prefix: {245: stats_intro, 249: stats_puncture, 235: stats_deprecated}}

        self.assertEqual(0, self.stats_ep.get_message_received(self.prefix))
        self.assertEqual(3, self.stats_ep.get_message_received(self.prefix, True, True, True))
        self.assertEqual(1, self.stats_ep.get_message_received(self.prefix, include_introduction=True))
        self.assertEqual(1, self.stats_ep.get_message_received(self.prefix, include_puncture=True))
        self.assertEqual(1, self.stats_ep.get_message_received(self.prefix, include_deprecated=True))

    async def test_aggregate_statistics_sum(self):
        """
        Check if network stats are correctly summed.
        """
        stat1 = NetworkStat(1)
        stat1.num_up, stat1.num_down, stat1.bytes_up, stat1.bytes_down = 2, 3, 4, 5
        stat2 = NetworkStat(2)
        stat2.num_up, stat2.num_down, stat2.bytes_up, stat2.bytes_down = 6, 7, 8, 9
        self.stats_ep.statistics = {self.prefix: {stat1.identifier: stat1, stat2.identifier: stat2}}

        statistics = self.stats_ep.get_aggregate_statistics(self.prefix)

        self.assertIsNotNone(statistics)
        self.assertEqual(8, statistics['num_up'])
        self.assertEqual(10, statistics['num_down'])
        self.assertEqual(12, statistics['bytes_up'])
        self.assertEqual(14, statistics['bytes_down'])
        self.assertEqual(0, statistics['diff_time'])

    async def test_aggregate_statistics_diff_one(self):
        """
        Check if the time diff is correctly calculated, with one positive stat.
        """
        stat1 = NetworkStat(1)
        stat1.last_measured_down, stat1.last_measured_up = 0, 0
        stat1.first_measured_down, stat1.first_measured_up = 0, 0
        stat2 = NetworkStat(2)
        stat2.last_measured_down, stat2.last_measured_up = 0, 5
        stat2.first_measured_down, stat2.first_measured_up = 0, 1
        self.stats_ep.statistics = {self.prefix: {stat1.identifier: stat1, stat2.identifier: stat2}}

        statistics = self.stats_ep.get_aggregate_statistics(self.prefix)

        self.assertIsNotNone(statistics)
        self.assertEqual(4, statistics['diff_time'])

    async def test_aggregate_statistics_diff_two(self):
        """
        Check if the time diff is correctly calculated, with two positive stats.
        """
        stat1 = NetworkStat(1)
        stat1.last_measured_down, stat1.last_measured_up = 0, 6
        stat1.first_measured_down, stat1.first_measured_up = 0, 2
        stat2 = NetworkStat(2)
        stat2.last_measured_down, stat2.last_measured_up = 0, 5
        stat2.first_measured_down, stat2.first_measured_up = 0, 1
        self.stats_ep.statistics = {self.prefix: {stat1.identifier: stat1, stat2.identifier: stat2}}

        statistics = self.stats_ep.get_aggregate_statistics(self.prefix)

        self.assertIsNotNone(statistics)
        self.assertEqual(5, statistics['diff_time'])

    async def test_aggregate_statistics_diff_many(self):
        """
        Check if the time diff is correctly calculated, with four positive stats.
        """
        stat1 = NetworkStat(1)
        stat1.last_measured_down, stat1.last_measured_up = 7, 6
        stat1.first_measured_down, stat1.first_measured_up = 3, 2
        stat2 = NetworkStat(2)
        stat2.last_measured_down, stat2.last_measured_up = 0, 5
        stat2.first_measured_down, stat2.first_measured_up = 1, 2
        self.stats_ep.statistics = {self.prefix: {stat1.identifier: stat1, stat2.identifier: stat2}}

        statistics = self.stats_ep.get_aggregate_statistics(self.prefix)

        self.assertIsNotNone(statistics)
        self.assertEqual(6, statistics['diff_time'])



from .discovery import MockWalk
from .endpoint import AutoMockEndpoint
from ...dht.discovery import DHTDiscoveryCommunity
from ...keyvault.crypto import default_eccrypto
from ...messaging.interfaces.statistics_endpoint import StatisticsEndpoint
from ...peer import Peer
from ...peerdiscovery.network import Network
from ...util import maybe_coroutine


class MockIPv8(object):

    def __init__(self, crypto_curve, overlay_class, create_dht=False, enable_statistics=False,
                 *args, **kwargs):
        self.endpoint = AutoMockEndpoint()
        self.endpoint.open()

        # Enable statistics
        if enable_statistics:
            self.endpoint = StatisticsEndpoint(self, self.endpoint)

        self.network = Network()
        self.my_peer = Peer(default_eccrypto.generate_key(crypto_curve), self.endpoint.wan_address)

        # Load a DHT community if specified
        self.dht = None
        if create_dht:
            self.dht = DHTDiscoveryCommunity(self.my_peer, self.endpoint, self.network)
            kwargs.update({'dht': self.dht})

        self.overlay = overlay_class(self.my_peer, self.endpoint, self.network, *args, **kwargs)
        self.overlay._use_main_thread = False
        self.discovery = MockWalk(self.overlay)

        self.overlay.my_estimated_wan = self.endpoint.wan_address
        self.overlay.my_estimated_lan = self.endpoint.lan_address

        self.overlays = []
        self.strategies = []

        if enable_statistics:
            self.endpoint.enable_community_statistics(self.overlay.get_prefix(), True)

    def add_strategy(self, overlay, strategy, target_peers):
        self.overlays.append(overlay)
        self.strategies.append((strategy, target_peers))

    def get_overlay(self, overlay_cls):
        return next(self.get_overlays(overlay_cls), None)

    def get_overlays(self, overlay_cls):
        return (o for o in [self.dht, self.overlay] if isinstance(o, overlay_cls))

    def unload_overlay(self, instance):
        self.overlays = [overlay for overlay in self.overlays if overlay != instance]
        self.strategies = [(strategy, target_peers) for (strategy, target_peers) in self.strategies
                           if strategy.overlay != instance]
        return maybe_coroutine(instance.unload)

    async def stop(self, stop_loop=True):
        self.endpoint.close()
        await self.overlay.unload()
        if self.dht:
            await self.dht.unload()



from base64 import decodebytes

from ..base import TestBase
from ...keyvault.crypto import default_eccrypto
from ...keyvault.private.m2crypto import M2CryptoSK
from ...keyvault.public.m2crypto import M2CryptoPK


class TestSerialization(TestBase):
    """
    Test whether keys can be serialized and unserialized correctly.
    """

    def setUp(self):
        super(TestSerialization, self).setUp()
        self.ec = default_eccrypto
        self.key = self.ec.generate_key(u"very-low")
        self.key_nacl = self.ec.generate_key(u"curve25519")

    def test_private_to_bin(self):
        """
        Check if M2Crypto derived key bins are valid.
        """
        private_bin = self.key.key_to_bin()

        self.assertTrue(self.ec.is_valid_private_bin(private_bin))

    def test_private_nacl_to_bin(self):
        """
        Check if libnacl derived key bins are valid.
        """
        private_bin = self.key_nacl.key_to_bin()

        self.assertTrue(self.ec.is_valid_private_bin(private_bin))

    def test_private_to_pem(self):
        """
        Check if keys can be serialized and loaded correctly in PEM format.
        """
        private_pem = self.key.key_to_pem()

        # Convert the PEM to a DER keystring
        prefix = "-----BEGIN EC PRIVATE KEY-----\n"
        postfix = "-----END EC PRIVATE KEY-----\n"
        keystring = decodebytes(private_pem[len(prefix):-len(postfix)])

        # Reconstruct a key with this keystring
        key = M2CryptoSK(keystring=keystring)

        self.assertEqual(private_pem, key.key_to_pem())

    def test_public_to_bin(self):
        """
        Check if M2Crypto derived public key bins are valid.
        """
        public_bin = self.key.pub().key_to_bin()

        self.assertTrue(self.ec.is_valid_public_bin(public_bin))

    def test_public_nacl_to_bin(self):
        """
        Check if libnacl derived public key bins are valid.
        """
        public_bin = self.key_nacl.pub().key_to_bin()

        self.assertTrue(self.ec.is_valid_public_bin(public_bin))

    def test_public_to_pem(self):
        """
        Check if public keys can be serialized and loaded correctly in PEM format.
        """
        public_pem = self.key.pub().key_to_pem()

        # Convert the PEM to a DER keystring
        prefix = "-----BEGIN PUBLIC KEY-----\n"
        postfix = "-----END PUBLIC KEY-----\n"
        keystring = decodebytes(public_pem[len(prefix):-len(postfix)])

        # Reconstruct a key with this keystring
        key = M2CryptoPK(keystring=keystring)

        self.assertEqual(public_pem, key.key_to_pem())



import typing

Address = typing.Tuple[str, int]

# pylint: disable=unused-import

if typing.TYPE_CHECKING:
    from ipv8.attestation.identity.attestation import Attestation
    from ipv8.attestation.identity.database import Credential
    from ipv8.attestation.identity.manager import PseudonymManager
    from ipv8.attestation.identity.metadata import Metadata
    from ipv8.attestation.identity_formats import IdentityAlgorithm
    from ipv8.attestation.tokentree.token import Token
    from ipv8.attestation.wallet.community import AttestationCommunity
    from ipv8.community import Community
    from ipv8.configuration import ConfigBuilder
    from ipv8.database import Database
    from ipv8.keyvault.keys import Key, PrivateKey, PublicKey
    from ipv8.messaging.interfaces.endpoint import Endpoint
    from ipv8.messaging.serialization import Payload
    from ipv8.overlay import Overlay
    from ipv8.peer import Peer
    from ipv8.requestcache import NumberCache
    from ipv8_service import IPv8

    IdentityAlgorithmClass = typing.Type[IdentityAlgorithm]
else:
    Attestation = 'ipv8.attestation.identity.attestation.Attestation'
    AttestationCommunity = 'ipv8.attestation.wallet.community.AttestationCommunity'
    Community = 'ipv8.community.Community'
    ConfigBuilder = 'ipv8.configuration.ConfigBuilder'
    Credential = 'ipv8.attestation.identity.database.Credential'
    Database = 'ipv8.database.Database'
    Endpoint = 'ipv8.messaging.interfaces.endpoint.Endpoint'
    IdentityAlgorithm = 'ipv8.attestation.identity_formats.IdentityAlgorithm'
    IdentityAlgorithmClass = 'ipv8.attestation.identity_formats.IdentityAlgorithm.__class__'
    IPv8 = 'ipv8_service.IPv8'
    Key = 'ipv8.keyvault.keys.Key'
    Metadata = 'ipv8.attestation.identity.metadata.Metadata'
    NumberCache = 'ipv8.requestcache.NumberCache'
    Overlay = 'ipv8.overlay.Overlay'
    Payload = 'ipv8.messaging.serialization.Payload'
    Peer = 'ipv8.peer.Peer'
    PrivateKey = 'ipv8.keyvault.keys.PrivateKey'
    PseudonymManager = 'ipv8.attestation.identity.manager.PseudonymManager'
    PublicKey = 'ipv8.keyvault.keys.PublicKey'
    Token = 'ipv8.attestation.tokentree.token.Token'



from ...messaging.payload import Payload


class RequestAttestationPayload(Payload):
    """
    Request an attestation based on some meta data.
    """
    msg_id = 5
    format_list = ['raw']

    def __init__(self, metadata):
        super(RequestAttestationPayload, self).__init__()
        self.metadata = metadata

    def to_pack_list(self):
        data = [('raw', self.metadata)]
        return data

    @classmethod
    def from_unpack_list(cls, metadata):
        return cls(metadata)


class VerifyAttestationRequestPayload(Payload):
    """
    Request an attestation by hash (published with metadata somewhere).
    """
    msg_id = 1
    format_list = ['20s']

    def __init__(self, hash):
        super(VerifyAttestationRequestPayload, self).__init__()
        self.hash = hash

    def to_pack_list(self):
        data = [('20s', self.hash)]
        return data

    @classmethod
    def from_unpack_list(cls, hash):
        return cls(hash)


class AttestationChunkPayload(Payload):
    """
    A chunk of Attestation.
    """
    msg_id = 2
    format_list = ['20s', 'H', 'raw']

    def __init__(self, hash, sequence_number, data):
        super(AttestationChunkPayload, self).__init__()
        self.hash = hash
        self.sequence_number = sequence_number
        self.data = data

    def to_pack_list(self):
        data = [('20s', self.hash),
                ('H', self.sequence_number),
                ('raw', self.data)]

        return data

    @classmethod
    def from_unpack_list(cls, hash, sequence_number, data):
        return cls(hash, sequence_number, data)


class ChallengePayload(Payload):
    """
    A challenge for an Attestee by a Verifier
    """
    msg_id = 3
    format_list = ['20s', 'raw']

    def __init__(self, attestation_hash, challenge):
        self.attestation_hash = attestation_hash
        self.challenge = challenge

    def to_pack_list(self):
        data = [('20s', self.attestation_hash),
                ('raw', self.challenge)]
        return data

    @classmethod
    def from_unpack_list(cls, attestation_hash, challenge):
        return cls(attestation_hash, challenge)


class ChallengeResponsePayload(Payload):
    """
    A challenge response from an Attestee to a Verifier
    """
    msg_id = 4
    format_list = ['20s', 'raw']

    def __init__(self, challenge_hash, response):
        self.challenge_hash = challenge_hash
        self.response = response

    def to_pack_list(self):
        data = [('20s', self.challenge_hash),
                ('raw', self.response)]
        return data

    @classmethod
    def from_unpack_list(cls, challenge_hash, response):
        return cls(challenge_hash, response)



"""
Additional dependencies:

python3 -m pip install PyQtWebEngine

To run:
export PYTHONPATH=.
python3 -m ipv8.attestation.wallet.irmaexact.enroll_script
"""

import base64
import binascii
import json
import sys
import time

from PyQt5.QtCore import *
from PyQt5.QtWebEngineWidgets import *
from PyQt5.QtWidgets import QApplication

import urllib3

from .gabi.attributes import make_attribute_list
from .gabi.builder import BuildDistributedProofList, Challenge, CredentialBuilder, IssueCommitmentMessage, IssueSignatureMessage
from .gabi.keys import CLSignature, DefaultSystemParameters
from .gabi.proofs import ProofP, ProofPCommitment, ProofS
from .keydump import nijmegen_pk_1623505755 as nijmegen_pk
from .wrappers import serialize_proof_d
from ..primitives.cryptography_wrapper import generate_safe_prime
from ....util import int2byte

my_app = QApplication(sys.argv)
my_web = QWebEngineView()
profile = QWebEngineProfile("storage", my_web)
cookie_store = profile.cookieStore()
cookie_store.deleteAllCookies()

token = None


def print_result(r):
    global token
    if r:
        u = json.loads(r)["u"].encode()
        token = u.split(b'/')[-1].decode()
        my_web.hide()
        my_app.quit()


def page_loaded(ok):
    if ok and my_web.url() == QUrl('https://services.nijmegen.nl/irma/gemeente/issue?'):
        my_web.page().runJavaScript("window.irmaSessionPtr", print_result)


my_web.loadFinished.connect(page_loaded)
my_web.load(QUrl("https://services.nijmegen.nl/irma/gemeente/issue"))
my_web.setWindowTitle("Annoying Pop-up")
my_web.show()

my_app.exec_()

secret = generate_safe_prime(DefaultSystemParameters[1024].Lm)

no_pin = 'AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA=\\n'


def do_get(u):
    http = urllib3.PoolManager()
    headers = {'Content-Type': 'application/json',
               "X-IRMA-MinProtocolVersion": "2.4",
               "X-IRMA-MaxProtocolVersion": "2.4"}

    r = http.request('GET', u, headers=headers)
    try:
        out = json.loads(r.data)
    except ValueError:
        out = r.data

    return out


def do_post(u, content, extra_headers={}):
    http = urllib3.PoolManager()
    headers = {'Content-Type': 'application/json'}
    headers.update(extra_headers)

    r = http.request('POST', u, headers=headers, body=content)
    try:
        out = json.loads(r.data)
    except ValueError:
        out = r.data

    return out


u = "https://keyshare.privacybydesign.foundation/tomcat/irma_keyshare_server/api/v1/client/register"
u = do_post(u, "{\"username\":\"\",\"pin\":\"%s\",\"email\":null,\"language\":\"en\"}" % no_pin)['u']
response = do_get(u)
context = int(binascii.hexlify(base64.b64decode(response["context"])), 16)
nonce = int(binascii.hexlify(base64.b64decode(response["nonce"])), 16)

irmaid = response["credentials"][0]["attributes"]["pseudonym"]
u = "https://keyshare.privacybydesign.foundation/tomcat/irma_keyshare_server/api/v1/users/verify/pin"
jwt = do_post(u, "{\"id\":\"%s\",\"pin\":\"%s\"}" % (irmaid, no_pin))["message"]

extra_headers = {"Authorization": jwt, "X-IRMA-Keyshare-Username": ""}
u = "https://keyshare.privacybydesign.foundation/tomcat/irma_keyshare_server/api/v1/prove/getCommitments"
response = do_post(u, "["
                      "{\"issuer\":{\"identifier\":\"pbdf.gemeente\"},\"counter\": 1},"
                      "{\"issuer\":{\"identifier\":\"pbdf.gemeente\"},\"counter\": 1}"
                      "]", extra_headers=extra_headers)
Pcommit = response['c'][0][1]['Pcommit']
P = response['c'][0][1]['P']

# Get nijmegen issuance
u = 'https://gw.nijmegen.nl/irma/session/%s/' % token
issuance_output = do_get(u)
context = int(binascii.hexlify(base64.b64decode(issuance_output['context'])), 16)
nonce1 = int(binascii.hexlify(base64.b64decode(issuance_output['nonce'])), 16)

u = "https://keyshare.privacybydesign.foundation/tomcat/irma_keyshare_server/api/v1/prove/getResponse"
# 2 because: 'pbdf.gemeente.address' and 'pbdf.gemeente.personalData'
cbs = [CredentialBuilder(nijmegen_pk, context, secret, nonce1) for _ in range(2)]
for cb in cbs:
    cb.MergeProofPCommitment(ProofPCommitment(P, Pcommit))
challenge = Challenge(cbs, context, nonce1, False)
commit_jwt = do_post(u, str(challenge), extra_headers=extra_headers)

proofP_d = commit_jwt.split(b'.')[1]
lens = len(proofP_d)
lenx = lens - (lens % 4 if lens % 4 else 4)
proofP_d = json.loads(base64.decodebytes(proofP_d[:lenx]) + b'}')["ProofP"]
P = proofP_d["P"]
c = proofP_d["c"]
s_response = proofP_d["s_response"]
proofP = ProofP(P, c, s_response)

u = 'https://gw.nijmegen.nl/irma/session/%s/status' % token

while do_get(u) != "CONNECTED":
    time.sleep(0.5)

u = 'https://gw.nijmegen.nl/irma/session/%s/commitments' % token

proofs = BuildDistributedProofList(cbs, challenge, [])
commitMsg = IssueCommitmentMessage(None, proofs, nonce1)


def proof_to_str(proof):
    return ('{"U": ' + str(proof.U)
            + ', "c": ' + str(proof.C)
            + ', "v_prime_response": ' + str(proof.VPrimeResponse)
            + ', "s_response": ' + str(proof.SResponse) + '}')


outjson = '{'
outjson += '"U": null,'
outjson += '"combinedProofs": [' + ', '.join(proof_to_str(p) for p in commitMsg.Proofs) + '],'
outjson += '"indices": [],'
outjson += '"n_2": ' + str(commitMsg.Nonce2) + ","
outjson += '"proofPJwt": "",'
outjson += '"proofPJwts": {"pbdf": "' + commit_jwt.decode() + '"}'
outjson += '}'

output_proof = do_post(u, outjson)


def b64_to_int(s):
    return str_to_int(base64.b64decode(s))


def str_to_int(s):
    if isinstance(s, str):
        s = b''.join(int2byte(ord(c)) for c in s)
    if s is None:
        return 0
    if s == b"":
        return 1
    return int(binascii.hexlify(s), 16)


isms = []
for ism in output_proof:
    proof_s_desc = ism["proof"]
    sig_desc = ism["signature"]
    proof_s = ProofS(b64_to_int(proof_s_desc["c"]), b64_to_int(proof_s_desc["e_response"]))
    signature = CLSignature(b64_to_int(sig_desc["A"]), b64_to_int(sig_desc["e"]), b64_to_int(sig_desc["v"]))
    isms.append(IssueSignatureMessage(signature, proof_s))


order_map = {
    'pbdf.gemeente.address': ["street", "houseNumber", "zipcode", "municipality", "city"],
    'pbdf.gemeente.personalData': ["initials", "firstnames", "prefix", "familyname", "fullname", "gender",
                                   "nationality", "surname", "dateofbirth", "cityofbirth", "countryofbirth", "over12",
                                   "over16", "over18", "over21", "over65", "bsn", "digidlevel"]
}


for i in range(len(cbs)):
    cb = cbs[i]
    ism = isms[i]
    ordering = order_map[issuance_output["credentials"][i]['credential']]
    attribute_ints, signing_date = make_attribute_list(issuance_output["credentials"][i], ordering)

    credential = cb.ConstructCredential(ism, attribute_ints)
    builder = credential.CreateDisclosureProofBuilder(list(range(1, len(attribute_ints) + 1)))
    builder.MergeProofPCommitment(ProofPCommitment(P, Pcommit))
    commit_randomizer = generate_safe_prime(nijmegen_pk.Params.LmCommit)
    A, Z = builder.Commit(commit_randomizer)
    p = builder.CreateProof(challenge)
    p.MergeProofP(proofP, nijmegen_pk)

    attr_output = '{\n'
    attr_output += '\t"sign_date": ' + str(signing_date) + ',\n'
    attr_output += '\t"proofd": "' + binascii.hexlify(serialize_proof_d(p)).decode() + '",\n'
    attr_output += '\t"z": ' + str(Z) + '\n'
    attr_output += '}'

    print("*" * 20)  # noqa: T001
    print("Attribute:", issuance_output["credentials"][i]['credential'])  # noqa: T001
    print(attr_output)  # noqa: T001
    print("*" * 20)  # noqa: T001



import logging
from collections import OrderedDict
from hashlib import sha3_256
from typing import Dict, List, Optional, Set

from .token import Token
from ...types import PrivateKey, PublicKey


class TokenTree(object):
    """
    Raw datatype for chains of double pointers (Tokens).

    It takes the following form:

        SHA3-256(PUBLIC KEY) <- TOKEN <- TOKEN <- ...

    Each token also has a SHA3-256 pointer to external content.

    Note that the public key has to be known to fulfil the genesis pointer to the SHA3-256 hash (the public key is not
    stored in the Tokens themselves).
    """

    def __init__(self, public_key: Optional[PublicKey] = None, private_key: Optional[PrivateKey] = None) -> None:
        """
        Create a new view of another's chain by specifying a public key or create your own chain by supplying
        a private key.

        :param public_key: the public key of the owner of this chain.
        :param private_key: the private key to use to add tokens to this chain.
        """
        super(TokenTree, self).__init__()
        self._logger = logging.getLogger(self.__class__.__name__)

        self.elements: Dict[bytes, Token] = {}
        self.unchained: OrderedDict = OrderedDict()
        self.unchained_max_size = 100

        if public_key is not None and private_key is None:
            self.public_key = public_key.pub()
            self.private_key = None
        elif public_key is None and private_key is not None:
            self.private_key = private_key
            self.public_key = private_key.pub()
        else:
            raise RuntimeError("Specify either public_key or private_key!")

        self.genesis_hash = sha3_256(self.public_key.key_to_bin()).digest()

    def add(self, content: bytes, after: Optional[Token] = None) -> Token:
        """
        Tokenize new content and add it to this chain.

        :param content: the content to tokenize and add to this chain.
        :param after: the token to succeed.
        :returns: the newly added Token.
        """
        if self.private_key is None:
            raise RuntimeError("Attempted to create token without a key!")
        previous_hash = self.genesis_hash if not after else after.get_hash()
        return self._append(Token(previous_hash, content=content, private_key=self.private_key))

    def add_by_hash(self, content_hash: bytes, after: Optional[Token] = None) -> Token:
        """
        Add the promise of tokenized content to this chain.

        :param content_hash: the hash of the content to add to this chain.
        :param after: the token to succeed.
        :returns: the newly added Token.
        """
        if self.private_key is None:
            raise RuntimeError("Attempted to create token without a key!")
        previous_hash = self.genesis_hash if not after else after.get_hash()
        return self._append(Token(previous_hash, content_hash=content_hash, private_key=self.private_key))

    def gather_token(self, token: Token) -> Optional[Token]:
        """
        Attempt to add received data to this chain.
        Data may be pending missing Tokens before being added to the chain structure.

        :param token: the token to attempt to add.
        :returns: the newly added token or None if the operation was not successful.
        """
        if token.verify(self.public_key):
            if token.previous_token_hash != self.genesis_hash and token.previous_token_hash not in self.elements:
                self.unchained[token] = None
                if len(self.unchained) > self.unchained_max_size:
                    self.unchained.popitem(False)
                self._logger.info(f"Delaying unchained token {token}!")
                return None
            elif token.get_hash() in self.elements:
                shadow_token = self.elements[token.get_hash()]
                if shadow_token.content is None and token.content is not None:
                    shadow_token.receive_content(token.content)
                return shadow_token
            else:
                self._append_chain_reaction_token(token)
            return token
        return None

    def get_missing(self) -> Set[bytes]:
        """
        Gather all the preceding hashes that have been specified but not collected.

        :returns: the set of missing hashes.
        """
        return {token.previous_token_hash for token in self.unchained}

    def verify(self, token: Token, maxdepth: int = 1000) -> bool:
        """
        Verify the chain integrity: all preceding tokens are correctly signed and stored.

        Should only be done when loading from database or direct writing (don't do that) by programmers.

        :param token: the token to start checking from.
        :param maxdepth: the maximum amount of steps to verify (after which this returns False).
        :returns: whether all preceding tokens were correct and signed.
        """
        current = token
        steps = 0
        while maxdepth == -1 or maxdepth > steps:
            if not current.verify(self.public_key):
                return False
            if current.previous_token_hash == self.genesis_hash:
                break
            if current.previous_token_hash not in self.elements:
                return False
            current = self.elements[current.previous_token_hash]
            steps += 1
        return steps < maxdepth

    def get_root_path(self, token: Token, maxdepth: int = 1000) -> List[Token]:
        """
        Calculate the path back to the root, including this token.

        :param token: the token to start checking from.
        :param maxdepth: the maximum amount of steps (after which this returns an empty list).
        :returns: the length of the path back to the root or an empty list if it doesn't exist.
        """
        current = token
        steps = 0
        path = [token]
        while maxdepth == -1 or maxdepth > steps:
            if not current.verify(self.public_key):
                return []
            if current.previous_token_hash == self.genesis_hash:
                break
            if current.previous_token_hash not in self.elements:
                return []
            current = self.elements[current.previous_token_hash]
            path += [current]
            steps += 1
        if steps < maxdepth:
            return path
        else:
            return []

    def serialize_public(self, up_to: Optional[Token] = None) -> bytes:
        """
        Serialize all the signed double pointers of this chain.

        :param up_to: the token to work back from to the root of the tree.
        """
        if up_to:
            # End specified, move back to the root
            out = up_to.get_plaintext_signed()
            next_token = up_to.previous_token_hash
            while next_token in self.elements:
                token = self.elements[next_token]
                out += token.get_plaintext_signed()
                next_token = token.previous_token_hash
            return out
        else:
            # Do the full tree dump.
            return b''.join(token.get_plaintext_signed() for token in self.elements.values())

    def unserialize_public(self, s: bytes) -> bool:
        """
        Given a serialized tree format, unserialize with the tree's public key.

        :param s: the serialized tree data.
        :returns: if all information was correctly unserialized.
        """
        sig_len = self.public_key.get_signature_length()
        chunk_size = 64 + sig_len
        correct = True
        for i in range(0, len(s), chunk_size):
            correct &= self.gather_token(Token.unserialize(s, self.public_key, offset=i)) is not None
        return correct

    def _append(self, token: Token) -> Token:
        """
        Append a token to this tree. Never call this directly: use add() instead!

        :param token: the token to append to the chain.
        :returns: the appended token.
        """
        self.elements[token.get_hash()] = token
        return token

    def _append_chain_reaction_token(self, token: Token) -> None:
        """
        Append the given token and wake up any tokens stored in `unchained` that pointed to it.

        :param token: the token to append to the chain.
        """
        self._append(token)
        retry_token = None
        for lost_token in self.unchained:
            if lost_token.previous_token_hash == token.get_hash():
                retry_token = lost_token
                break
        if retry_token is not None:
            self.unchained.pop(retry_token)
            if self.gather_token(retry_token) is None:
                self._logger.warning(f"Dropped illegal token {retry_token}!")



import base64
import random

from ..REST.rest_base import RESTTestBase
from ...keyvault.crypto import default_eccrypto
from ...peer import Peer


class TestNetworkEndpoint(RESTTestBase):

    mock_peer = Peer(default_eccrypto.generate_key("curve25519"), address=("1.2.3.4", 5))
    mock_b64mid = base64.b64encode(mock_peer.mid).decode()
    mock_b64pubkey = base64.b64encode(mock_peer.public_key.key_to_bin()).decode()

    async def setUp(self):
        super(TestNetworkEndpoint, self).setUp()
        await self.initialize([], 1)
        self.ipv8 = self.node(0)

    async def test_no_peers(self):
        """
        Check if the network endpoint returns no peers if it has no peers.
        """
        peer_response = await self.make_request(self.ipv8, "network", "GET")

        self.assertIn("peers", peer_response)
        self.assertDictEqual({}, peer_response["peers"])

    async def test_one_peer_no_services(self):
        """
        Check if the network endpoint returns its one known peer with no services.
        """
        self.ipv8.network.add_verified_peer(self.mock_peer)

        peer_response = await self.make_request(self.ipv8, "network", "GET")

        self.assertIn("peers", peer_response)
        self.assertIn(self.mock_b64mid, peer_response["peers"])
        self.assertEqual("1.2.3.4", peer_response["peers"][self.mock_b64mid]["ip"])
        self.assertEqual(5, peer_response["peers"][self.mock_b64mid]["port"])
        self.assertEqual(self.mock_b64pubkey, peer_response["peers"][self.mock_b64mid]["public_key"])
        self.assertListEqual([], peer_response["peers"][self.mock_b64mid]["services"])

    async def test_one_peer_one_service(self):
        """
        Check if the network endpoint returns its one known peer with one service.
        """
        self.ipv8.network.add_verified_peer(self.mock_peer)
        mock_service = bytes(list(range(20)))
        mock_b64service = base64.b64encode(mock_service).decode()
        self.ipv8.network.discover_services(self.mock_peer, [mock_service])

        peer_response = await self.make_request(self.ipv8, "network", "GET")

        self.assertIn("peers", peer_response)
        self.assertIn(self.mock_b64mid, peer_response["peers"])
        self.assertEqual("1.2.3.4", peer_response["peers"][self.mock_b64mid]["ip"])
        self.assertEqual(5, peer_response["peers"][self.mock_b64mid]["port"])
        self.assertEqual(self.mock_b64pubkey, peer_response["peers"][self.mock_b64mid]["public_key"])
        self.assertListEqual([mock_b64service], peer_response["peers"][self.mock_b64mid]["services"])

    async def test_one_peer_multiple_services(self):
        """
        Check if the network endpoint returns its one known peer with multiple services.
        """
        self.ipv8.network.add_verified_peer(self.mock_peer)
        mock_services = [bytes([random.randint(0, 255) for _ in range(20)]) for __ in range(30)]
        mock_b64services = [base64.b64encode(mock_service).decode() for mock_service in mock_services]
        self.ipv8.network.discover_services(self.mock_peer, mock_services)

        peer_response = await self.make_request(self.ipv8, "network", "GET")

        self.assertIn("peers", peer_response)
        self.assertIn(self.mock_b64mid, peer_response["peers"])
        self.assertEqual("1.2.3.4", peer_response["peers"][self.mock_b64mid]["ip"])
        self.assertEqual(5, peer_response["peers"][self.mock_b64mid]["port"])
        self.assertEqual(self.mock_b64pubkey, peer_response["peers"][self.mock_b64mid]["public_key"])
        self.assertSetEqual(set(mock_b64services), set(peer_response["peers"][self.mock_b64mid]["services"]))

    async def test_multiple_peers_multiple_services(self):
        """
        Check if the network endpoint returns multiple peers with distinct services.
        """
        mock_peers = [Peer(default_eccrypto.generate_key("curve25519")) for _ in range(20)]
        mock_b64services = {}
        for mock_peer in mock_peers:
            b64mid = base64.b64encode(mock_peer.mid).decode()
            self.ipv8.network.add_verified_peer(mock_peer)
            local_services = [bytes([random.randint(0, 255) for _ in range(20)]) for __ in range(30)]
            mock_b64services[b64mid] = [base64.b64encode(s).decode() for s in local_services]
            self.ipv8.network.discover_services(mock_peer, local_services)

        peer_response = await self.make_request(self.ipv8, "network", "GET")

        self.assertIn("peers", peer_response)
        for b64mid, peer_descriptor in peer_response["peers"].items():
            self.assertIn(b64mid, [base64.b64encode(p.mid).decode() for p in mock_peers])
            self.assertIn(peer_descriptor["public_key"],
                          [base64.b64encode(p.public_key.key_to_bin()).decode() for p in mock_peers])
            self.assertEqual("0.0.0.0", peer_descriptor["ip"])
            self.assertEqual(0, peer_descriptor["port"])
            self.assertSetEqual(set(mock_b64services[b64mid]), set(peer_descriptor["services"]))



from ....base import TestBase
from .....attestation.default_identity_formats import FORMATS
from .....attestation.wallet.irmaexact.algorithm import IRMAExactAlgorithm


class TestAlgorithm(TestBase):

    def setUp(self):
        super(TestAlgorithm, self).setUp()
        self.blob = """
        {
            "sign_date": 2586,
            "proofd": "0120ae8c08383eb6326ed4d690472c14167157ba95bddfc6eb961475d11146c78b7702010014ca54a5ca54059afa2ef17b9a8826082f01dc75a58cae5ec18a3b3f93b5de357dbb5b7a326f9ffb67aef043f7706dddd272c9b5e46d48598975bbb22ad28fbeda52e1384fc39edbfae6d48d6cb03148dd0b3bc7cf13b594529c3d9c9f325e847add7d8cf1007e8f8fb5cc4cedca88e9f4ebcc69eba22741e8ede9744677be96df68ad0f689052264e8ddd67f15b89ca106f9424efe0936fc3ea567573900a39254c28b9fcb4c805de6a55d13b716aac4669ed9f4620e98ed54012f4dfa633769e510961296e228f5152c1152b760ff58ebe964dd84bf9a6b39da9b4acee1cedcab3576d95dcffbd9cdea6bf61726f0468b2f405e825103bcb20763d0f69decb013f2425ecd45cf8d8a8e6feabcb2c7dbddce283c494fc309451dc310f3ca36de5e2bab532d50734bcbe250bcdae7cc7b6e978e2dfe5c4c11521aa4326b7683c91020191083b68a01f222fa9a6260d0b40bce3669a7dad64c9ec6bdb7418865f6c07aa616e06fbeb314f32f6f8d9773acc4cf96667044007a5e8f8364c67884219ddc84057f94f06c8c3677f8483da72a99ed05976856ca1597ecc3164d7f4236736b64b6b656e7bef4bfd30cfd51bcb4bef990ca0ff88b5b0cae9df789acb34dc6e34d3daf125e7ba13d34d70b3ba2085a85f1f5e5e39f704cd8af7787ec29f204b258abfe9fbfa644e79ac384f15817fb181255b57f1f8e3c6fae041422feb8786259356287266410ee4cf595418bcb392367ed17b8c3dbfd87c97d338dc919c99579294d301fcb54dcfde75f21ae310a390473e7801a9f5c339a861c50f5836175a92f31327c6b8fe82ff7bfb6a915cfe3a950765f5a4261ce1b22146435fe4aaf93e5a0e3e6febcc0542615dbfb0b1ab4bc09e3a1f7c4be574a268a5c9fd255cadf53d0924783965109f6de678210e6bd1ed66a3a11081e0958f20d01860208d1d2b7419576131e4e5759729defb2eceb0ec413fa2c1bb76a3f60fb1c41397c2d50f127be92d3dfc42bfd44d68dee28ce0617c0150d945cc7a583966ae5dc8e8340c24d2004eeaef1e073c6d2e71a55eb0e0cb6ffa56dc9f217631445bb8157548f7ddc735bb703f89e644d4ac6a5c5453312dae28b4c184e4b089f3c18ceec147f844b940",
            "z": 21025909574607652013242068407286517850424307678122890876135156101653321984092105929057125603368801791220668605323274805950534554333885095151891554345408081373199060099202339143424919643961351117275291540867055856626869631650968369803183902895146289487368006719436569169044553294877369014424433240847941963958916265623517330529414890418453511520187847786782916076948797140109695922159477047382915875554999783288589498520057591077035129853458855982695920607696398615002875185515097109401895544388882159438043913551570342780281962902283409019706410544114516101827596857844194407086012298321703512788196253222325858419223
        }
        """

    def test_integration(self):
        id_format = "id_irma_nijmegen_ageLimits_1568208470"
        algorithm = IRMAExactAlgorithm(id_format, FORMATS)
        attestation_blob, _ = algorithm.import_blob(self.blob)
        attestation = algorithm.get_attestation_class().unserialize_private(None, attestation_blob, id_format)
        attestation_public = attestation.unserialize(attestation.serialize(), id_format)

        aggregate = algorithm.create_certainty_aggregate(attestation_public)
        challenges = algorithm.create_challenges(None, attestation_public)
        for challenge in challenges:
            response = algorithm.create_challenge_response(None, attestation, challenge)
            algorithm.process_challenge_response(aggregate, challenge, response)

        value = '{"over16": "Yes", "over12": "Yes", "over21": "Yes", "over65": "No", "over18": "Yes"}'
        self.assertEqual(1.0, algorithm.certainty(value, aggregate))

    def test_integration_incomplete(self):
        id_format = "id_irma_nijmegen_ageLimits_1568208470"
        algorithm = IRMAExactAlgorithm(id_format, FORMATS)

        attestation_blob, _ = algorithm.import_blob(self.blob)
        attestation = algorithm.get_attestation_class().unserialize_private(None, attestation_blob, id_format)
        attestation_public = attestation.unserialize(attestation.serialize(), id_format)

        aggregate = algorithm.create_certainty_aggregate(attestation_public)
        challenges = algorithm.create_challenges(None, attestation_public)
        for challenge in challenges[:-1]:
            response = algorithm.create_challenge_response(None, attestation, challenge)
            algorithm.process_challenge_response(aggregate, challenge, response)

        value = '{"over16": "Yes", "over12": "Yes", "over21": "Yes", "over65": "No", "over18": "Yes"}'
        self.assertEqual((algorithm.challenge_count - 1.0) / len(challenges), algorithm.certainty(value, aggregate))

    def test_integration_wrong(self):
        id_format = "id_irma_nijmegen_ageLimits_1568208470"
        algorithm = IRMAExactAlgorithm(id_format, FORMATS)

        attestation_blob, _ = algorithm.import_blob(self.blob)
        attestation = algorithm.get_attestation_class().unserialize_private(None, attestation_blob, id_format)
        attestation_public = attestation.unserialize(attestation.serialize(), id_format)

        aggregate = algorithm.create_certainty_aggregate(attestation_public)
        challenges = algorithm.create_challenges(None, attestation_public)
        for challenge in challenges:
            response = algorithm.create_challenge_response(None, attestation, challenge)
            algorithm.process_challenge_response(aggregate, challenge, response)

        value = '{"over16": "Yes", "over12": "Yes", "over21": "Yes", "over65": "Yes", "over18": "Yes"}'
        self.assertEqual(0.0, algorithm.certainty(value, aggregate))



from collections import deque

from .tunnel import CIRCUIT_STATE_READY, PEER_FLAG_EXIT_IPV8


class TunnelEndpoint(object):

    def __init__(self, endpoint):
        self.endpoint = endpoint
        self.hops = 0
        self.tunnel_community = None
        self.settings = {}
        self.send_queue = deque(maxlen=100)

    def set_tunnel_community(self, tunnel_community, hops=1):
        self.tunnel_community = tunnel_community
        self.hops = hops

    def set_anonymity(self, prefix, enable):
        self.settings[prefix] = enable

    def send(self, address, packet):
        prefix = packet[:22]
        if not self.settings.get(prefix, False):
            self.endpoint.send(address, packet)
            return

        if self.tunnel_community:
            circuits = self.tunnel_community.find_circuits(exit_flags=[PEER_FLAG_EXIT_IPV8], hops=self.hops, state=None)
            circuit = circuits[0] if circuits else None
            if not circuit or circuit.state != CIRCUIT_STATE_READY:
                # Recreate tunnel when needed
                if not circuit:
                    self.tunnel_community.create_circuit(self.hops, exit_flags=[PEER_FLAG_EXIT_IPV8])
                self.send_queue.append((address, packet))
                return

            circuit_id = circuit.circuit_id
            self.tunnel_community.send_data(circuit.peer, circuit_id, address, ('0.0.0.0', 0), packet)

            # Any packets still need sending?
            while self.send_queue:
                address, packet = self.send_queue.popleft()
                self.tunnel_community.send_data(circuit.peer, circuit_id, address, ('0.0.0.0', 0), packet)

    def notify_listeners(self, packet, from_tunnel=False):
        for listener in self._listeners:
            # Anonymized communities should ignore traffic received from the socket
            # Non-anonymized communities should ignore traffic received from the TunnelCommunity
            if getattr(listener, 'anonymize', False) != from_tunnel:
                continue
            self._deliver_later(listener, packet)

    def __getattribute__(self, item):
        try:
            return object.__getattribute__(self, item)
        except AttributeError:
            try:
                return object.__getattribute__(self.endpoint, item)
            except AttributeError:
                return object.__getattribute__(self.endpoint.endpoint, item)



import logging
import sys
import time
from asyncio import CancelledError, ensure_future, gather, get_event_loop, sleep
from base64 import b64decode
from contextlib import suppress
from os.path import isfile
from threading import RLock
from traceback import format_exception

if hasattr(sys.modules['__main__'], "IPv8"):
    sys.modules[__name__] = sys.modules['__main__']
else:
    if __name__ == '__main__' or __name__ == 'ipv8_service':
        from ipv8.messaging.interfaces.statistics_endpoint import StatisticsEndpoint
        from ipv8.attestation.identity.community import IdentityCommunity
        from ipv8.attestation.wallet.community import AttestationCommunity
        from ipv8.bootstrapping.dispersy.bootstrapper import DispersyBootstrapper
        from ipv8.bootstrapping.udpbroadcast.bootstrapper import UDPBroadcastBootstrapper
        from ipv8.keyvault.crypto import default_eccrypto
        from ipv8.keyvault.private.m2crypto import M2CryptoSK
        from ipv8.messaging.anonymization.community import TunnelCommunity
        from ipv8.messaging.anonymization.endpoint import TunnelEndpoint
        from ipv8.messaging.anonymization.hidden_services import HiddenTunnelCommunity
        from ipv8.messaging.interfaces.dispatcher.endpoint import DispatcherEndpoint
        from ipv8.messaging.interfaces.udp.endpoint import UDPEndpoint
        from ipv8.peer import Peer
        from ipv8.peerdiscovery.community import DiscoveryCommunity
        from ipv8.peerdiscovery.discovery import EdgeWalk, RandomWalk
        from ipv8.peerdiscovery.network import Network
        from ipv8.dht.discovery import DHTDiscoveryCommunity
        from ipv8.util import maybe_coroutine
    else:
        from .ipv8.messaging.interfaces.statistics_endpoint import StatisticsEndpoint
        from .ipv8.attestation.identity.community import IdentityCommunity
        from .ipv8.attestation.wallet.community import AttestationCommunity
        from .ipv8.bootstrapping.dispersy.bootstrapper import DispersyBootstrapper
        from .ipv8.bootstrapping.udpbroadcast.bootstrapper import UDPBroadcastBootstrapper
        from .ipv8.keyvault.crypto import default_eccrypto
        from .ipv8.keyvault.private.m2crypto import M2CryptoSK
        from .ipv8.messaging.anonymization.community import TunnelCommunity
        from .ipv8.messaging.anonymization.endpoint import TunnelEndpoint
        from .ipv8.messaging.anonymization.hidden_services import HiddenTunnelCommunity
        from .ipv8.messaging.interfaces.dispatcher.endpoint import DispatcherEndpoint
        from .ipv8.messaging.interfaces.udp.endpoint import UDPEndpoint
        from .ipv8.peer import Peer
        from .ipv8.peerdiscovery.community import DiscoveryCommunity
        from .ipv8.peerdiscovery.discovery import EdgeWalk, RandomWalk
        from .ipv8.peerdiscovery.network import Network
        from .ipv8.dht.discovery import DHTDiscoveryCommunity
        from .ipv8.util import maybe_coroutine

    _COMMUNITIES = {
        'AttestationCommunity': AttestationCommunity,
        'DiscoveryCommunity': DiscoveryCommunity,
        'HiddenTunnelCommunity': HiddenTunnelCommunity,
        'IdentityCommunity': IdentityCommunity,
        'TunnelCommunity': TunnelCommunity,
        'DHTDiscoveryCommunity': DHTDiscoveryCommunity,
    }

    _WALKERS = {
        'EdgeWalk': EdgeWalk,
        'RandomWalk': RandomWalk
    }

    _BOOTSTRAPPERS = {
        'DispersyBootstrapper': DispersyBootstrapper,
        'UDPBroadcastBootstrapper': UDPBroadcastBootstrapper
    }

    class IPv8(object):

        def __init__(self, configuration, endpoint_override=None, enable_statistics=False, extra_communities=None):
            super(IPv8, self).__init__()
            self.configuration = configuration

            if endpoint_override:
                self.endpoint = endpoint_override
            else:
                self.endpoint = DispatcherEndpoint(["UDPIPv4"],
                                                   UDPIPv4={'port': configuration['port'],
                                                            'ip': configuration['address']})
            if enable_statistics:
                self.endpoint = StatisticsEndpoint(self.endpoint)
            if any([overlay.get('initialize', {}).get('anonymize') for overlay in configuration['overlays']]):
                self.endpoint = TunnelEndpoint(self.endpoint)

            self.network = Network()

            # Load/generate keys
            self.keys = {}
            for key_block in configuration['keys']:
                if key_block['file'] and isfile(key_block['file']):
                    with open(key_block['file'], 'rb') as f:
                        content = f.read()
                        try:
                            # IPv8 Standardized bin format
                            self.keys[key_block['alias']] = Peer(default_eccrypto.key_from_private_bin(content))
                        except ValueError:
                            try:
                                # Try old Tribler M2Crypto PEM format
                                content = b64decode(content[31:-30].replace('\n', ''))
                                peer = Peer(M2CryptoSK(keystring=content))
                                peer.mid  # This will error out if the keystring is not M2Crypto
                                self.keys[key_block['alias']] = peer
                            except Exception:
                                # Try old LibNacl format
                                content = "LibNaCLSK:" + content
                                self.keys[key_block['alias']] = Peer(default_eccrypto.key_from_private_bin(content))
                else:
                    self.keys[key_block['alias']] = Peer(default_eccrypto.generate_key(key_block['generation']))
                    if key_block['file']:
                        with open(key_block['file'], 'wb') as f:
                            f.write(self.keys[key_block['alias']].key.key_to_bin())

            # Setup logging
            logging.basicConfig(**configuration['logger'])

            self.overlay_lock = RLock()
            self.strategies = []
            self.overlays = []
            self.on_start = []

            for overlay in configuration['overlays']:
                overlay_class = _COMMUNITIES.get(overlay['class'], (extra_communities or {}).get(overlay['class']))
                my_peer = self.keys[overlay['key']]
                overlay_instance = overlay_class(my_peer, self.endpoint, self.network, **overlay['initialize'])
                self.overlays.append(overlay_instance)
                for walker in overlay['walkers']:
                    strategy_class = _WALKERS.get(walker['strategy'],
                                                  overlay_instance.get_available_strategies().get(walker['strategy']))
                    args = walker['init']
                    target_peers = walker['peers']
                    self.strategies.append((strategy_class(overlay_instance, **args), target_peers))
                for config in overlay['on_start']:
                    self.on_start.append((getattr(overlay_instance, config[0]), config[1:]))
                for bootstrapper in overlay['bootstrappers']:
                    bootstrapper_class = _BOOTSTRAPPERS.get(bootstrapper['class'])
                    if bootstrapper_class:
                        overlay_instance.bootstrappers.append(bootstrapper_class(**bootstrapper['init']))
            self.walk_interval = configuration['walker_interval']
            self.state_machine_task = None

        async def start(self):
            await self.endpoint.open()

            await gather(*(ensure_future(maybe_coroutine(func, *args)) for func, args in self.on_start))

            async def ticker():
                while True:
                    await self.on_tick()
                    await sleep(0.0)
            self.state_machine_task = ensure_future(ticker())

        async def on_tick(self):
            if self.endpoint.is_open():
                with self.overlay_lock:
                    smooth = self.walk_interval // len(self.strategies) if self.strategies else 0
                    ticker = len(self.strategies)
                    for strategy, target_peers in self.strategies:
                        peer_count = len(strategy.overlay.get_peers())
                        start_time = time.time()
                        if (target_peers == -1) or (peer_count < target_peers):
                            # We wrap the take_step into a general except as it is prone to programmer error.
                            try:
                                strategy.take_step()
                            except Exception:
                                logging.error("Exception occurred while trying to walk!\n"
                                              + ''.join(format_exception(*sys.exc_info())))
                        ticker -= 1 if ticker else 0
                        sleep_time = smooth - (time.time() - start_time)
                        if ticker and sleep_time > 0.01:
                            await sleep(sleep_time)
                        if self.state_machine_task.done():
                            # By awaiting, we might have been stopped.
                            # In that case, exit out of the loop.
                            break
                    else:
                        await sleep(self.walk_interval)

        def add_strategy(self, overlay, strategy, target_peers):
            with self.overlay_lock:
                if overlay not in self.overlays:
                    self.overlays.append(overlay)
                self.strategies.append((strategy, target_peers))

        def unload_overlay(self, instance):
            with self.overlay_lock:
                self.overlays = [overlay for overlay in self.overlays if overlay != instance]
                self.strategies = [(strategy, target_peers) for (strategy, target_peers) in self.strategies
                                   if strategy.overlay != instance]
                return maybe_coroutine(instance.unload)

        def get_overlay(self, overlay_cls):
            return next(self.get_overlays(overlay_cls), None)

        def get_overlays(self, overlay_cls):
            return (o for o in self.overlays if isinstance(o, overlay_cls))

        async def produce_anonymized_endpoint(self):
            base_endpoint = UDPEndpoint(port=0, ip=self.configuration['address'])
            await base_endpoint.open()
            return TunnelEndpoint(base_endpoint)

        async def stop(self, stop_loop=True):
            if self.state_machine_task:
                self.state_machine_task.cancel()
                with suppress(CancelledError):
                    await self.state_machine_task
            with self.overlay_lock:
                unload_list = [self.unload_overlay(overlay) for overlay in self.overlays[:]]
                await gather(*unload_list)
                self.endpoint.close()
            if stop_loop:
                loop = get_event_loop()
                loop.call_later(0, loop.stop)


if __name__ == '__main__':
    from scripts.ipv8_plugin import main
    main(sys.argv[1:])



import struct

from .attestation import (attest_sha256, attest_sha256_4, attest_sha512, binary_relativity_certainty,
                          binary_relativity_sha256, binary_relativity_sha256_4, binary_relativity_sha512,
                          create_challenge, create_challenge_response_from_pair, create_empty_relativity_map,
                          create_honesty_check, process_challenge_response)
from .structs import BonehAttestation
from ..primitives.boneh import generate_keypair
from ..primitives.structs import BonehPrivateKey, BonehPublicKey, pack_pair, unpack_pair
from ...identity_formats import IdentityAlgorithm


class BonehExactAlgorithm(IdentityAlgorithm):

    def __init__(self, id_format, formats):
        super(BonehExactAlgorithm, self).__init__(id_format, formats)
        self.honesty_check = True

        # Check algorithm match
        if formats[id_format]["algorithm"] != "bonehexact":
            raise RuntimeError("Identity format linked to wrong algorithm")

        # Check key size match
        self.key_size = formats[self.id_format]["key_size"]
        if self.key_size < 32 or self.key_size > 512:
            raise RuntimeError("Illegal key size specified")

        # Check hash mode match
        hash_mode = formats[self.id_format]["hash"]
        if hash_mode == "sha256":
            self.attest_function = attest_sha256
            self.aggregate_reference = binary_relativity_sha256
        elif hash_mode == "sha256_4":
            self.attest_function = attest_sha256_4
            self.aggregate_reference = binary_relativity_sha256_4
        elif hash_mode == "sha512":
            self.attest_function = attest_sha512
            self.aggregate_reference = binary_relativity_sha512
        else:
            raise RuntimeError("Unknown hashing mode")

    def generate_secret_key(self):
        """
        Generate a secret key.

        :return: the secret key
        """
        return generate_keypair(self.key_size)[1]

    def load_secret_key(self, serialized):
        """
        Unserialize a secret key from the key material.

        :param serialized: the string of the private key
        :return: the private key
        """
        return BonehPrivateKey.unserialize(serialized)

    def load_public_key(self, serialized):
        """
        Unserialize a public key from the key material.

        :param serialized: the string of the public key
        :return: the public key
        """
        return BonehPublicKey.unserialize(serialized)

    def get_attestation_class(self):
        """
        Return the Attestation (sub)class for serialization

        :return: the Attestation object
        :rtype: BonehAttestation
        """
        return BonehAttestation

    def attest(self, PK, value):
        """
        Attest to a value for a certain public key.

        :param PK: the public key of the party we are attesting for
        :type PK: BonehPublicKey
        :param value: the value we are attesting to
        :type value: str
        :return: the attestation string
        :rtype: str
        """
        return self.attest_function(PK, value).serialize()

    def certainty(self, value, aggregate):
        """
        The current certainty of the aggregate object representing a certain value.

        :param value: the value to match to
        :type value: str
        :param aggregate: the aggregate object
        :type aggregate: dict
        :return: the matching factor [0.0-1.0]
        :rtype: float
        """
        return binary_relativity_certainty(self.aggregate_reference(value), aggregate)

    def create_challenges(self, PK, attestation):
        """
        Create challenges for a certain counterparty.

        :param PK: the public key of the party we are challenging
        :type PK: BonehPublicKey
        :param attestation: the attestation information
        :type attestation: BonehAttestation
        :return: the challenges to send
        :rtype: [str]
        """
        challenges = []
        for bitpair in attestation.bitpairs:
            challenge = create_challenge(attestation.PK, bitpair)
            serialized = pack_pair(challenge.a, challenge.b)
            challenges.append(serialized)
        return challenges

    def create_challenge_response(self, SK, attestation, challenge):
        """
        Create an honest response to a challenge of our value.

        :param SK: our secret key
        :type SK: BonehPrivateKey
        :param attestation: the attestation information
        :type attestation: Attestation
        :param challenge: the challenge to respond to
        :return: the response to a challenge
        :rtype: str
        """
        return struct.pack(">B", create_challenge_response_from_pair(SK, unpack_pair(challenge)))

    def create_certainty_aggregate(self, attestation):
        """
        Create an empty aggregate object, for matching to values.

        :param attestation: the attestation information
        :type attestation: Attestation
        :return: the aggregate object
        :rtype: dict
        """
        return create_empty_relativity_map()

    def create_honesty_challenge(self, PK, value):
        """
        Use a known value to check for honesty.

        :param PK: the public key of the party we are challenging
        :type PK: BonehPublicKey
        :param value: the value to use
        :type value: str
        :return: the challenge to send
        :rtype: str
        """
        raw_challenge = create_honesty_check(PK, value)
        return pack_pair(raw_challenge.a, raw_challenge.b)

    def process_honesty_challenge(self, value, response):
        """
        Given a response, check if it matches the expected value.

        :param value: the expected value
        :type value: int
        :param response: the returned response
        :type response: str
        :return: if the value matches the response
        :rtype: bool
        """
        unpacked, = struct.unpack(">B", response)
        return value == unpacked

    def process_challenge_response(self, aggregate, challenge, response):
        """
        Given a response, update the current aggregate.

        :param aggregate: the aggregate object
        :type aggregate: dict
        :param challenge: the sent challenge
        :type challenge: str
        :param response: the response to introduce
        :type response: str
        :return: the new aggregate
        :rtype: dict
        """
        unpacked, = struct.unpack(">B", response)
        return process_challenge_response(aggregate, unpacked)


__all__ = ["BonehAttestation", "BonehExactAlgorithm"]



from abc import ABC, abstractmethod
from asyncio import Future
from typing import Coroutine, Iterable, Union

from ..types import Address, Community


class Bootstrapper(ABC):
    """
    A script to connect to external bootstrapping resources, using external hardware.

    When not to crash:
     - Loading this module file when missing dependencies.
     - Initializing a subclass when the service is unavailable (return ``False``).
     - Failing to retrieve peers for a service (return an empty iterable, e.g. ``[]``).

    When to crash:
     - Initializing this module subclass when missing dependencies.
    """

    @abstractmethod
    def initialize(self, overlay: Community) -> Union[Future, Coroutine]:
        """
        Start connecting to this bootstrapping service. Don't perform any network traffic in ``__init__``!

        You are encourages to implement this method as non-async to have faster bootstrapper inclusion.

        :param overlay: the network overlay to initialize for.
        :returns: whether the initialization was successful.
        """

    @abstractmethod
    async def get_addresses(self, overlay: Community, timeout: float) -> Iterable[Address]:
        """
        Return some IPv8 addresses (if available) from this bootstrapping service.
        These addresses should be walkable (not blocked by a NAT or firewall).

        :param overlay: the network overlay to get peers for.
        :param timeout: the maximum time we wish to wait until we get any result (i.e. an empty list).
        :returns: the addresses for the given service_id.
        """

    @abstractmethod
    def keep_alive(self, overlay: Community) -> None:
        """
        Periodically called to keep this bootstrap connection alive.

        :param overlay: the network overlay to keep alive.
        """

    @abstractmethod
    def blacklist(self) -> Iterable[Address]:
        """
        Returns the blacklisted addresses for this Bootstrapper.
        """

    @abstractmethod
    def unload(self) -> None:
        """
        Stop and unload all the resources used by this Bootstrapper.
        """



from asyncio import Future
from unittest.mock import Mock

from .mock import MockDHTProvider
from ...base import TestBase
from ...mocking.endpoint import MockEndpointListener
from ...mocking.exit_socket import MockTunnelExitSocket
from ...mocking.ipv8 import MockIPv8
from ....messaging.anonymization.community import TunnelCommunity, TunnelSettings
from ....messaging.anonymization.endpoint import TunnelEndpoint
from ....messaging.anonymization.tunnel import (CIRCUIT_STATE_EXTENDING, PEER_FLAG_EXIT_BT,
                                                PEER_FLAG_EXIT_IPV8, PEER_FLAG_SPEED_TEST)
from ....messaging.interfaces.udp.endpoint import DomainAddress, UDPEndpoint
from ....util import succeed


class TestTunnelCommunity(TestBase):

    def setUp(self):
        super(TestTunnelCommunity, self).setUp()
        self.initialize(TunnelCommunity, 2)

        # An actual UDPEndpoint, if needed by the test (for catching exited data)
        self.public_endpoint = None

    async def tearDown(self):
        # If an endpoint was used, close it
        if self.public_endpoint:
            self.public_endpoint.close()

        return await super(TestTunnelCommunity, self).tearDown()

    def create_node(self):
        # Initialize a TunnelCommunity without circuits or exit node functionality
        settings = TunnelSettings()
        settings.min_circuits = 0
        settings.max_circuits = 0
        settings.remove_tunnel_delay = 0
        ipv8 = MockIPv8(u"curve25519", TunnelCommunity, settings=settings)
        # Then kill all automated circuit creation
        ipv8.overlay.cancel_all_pending_tasks()
        # Finally, use the proper exitnode and circuit settings for manual creation
        ipv8.overlay.settings.min_circuits = 1
        ipv8.overlay.settings.max_circuits = 1
        ipv8.overlay.dht_provider = MockDHTProvider(ipv8.overlay.my_peer)
        return ipv8

    def assert_no_more_tunnels(self):
        """
        Utility method to check whether there are no more tunnels left
        """
        for node in self.nodes:
            self.assertFalse(node.overlay.exit_sockets)
            self.assertFalse(node.overlay.relay_from_to)
            self.assertFalse(node.overlay.circuits)

    def settings(self, i):
        return self.overlay(i).settings

    async def test_introduction_as_exit(self):
        """
        Check if introduction requests share the fact that nodes are exit nodes.
        """
        self.overlay(0).settings.peer_flags.add(PEER_FLAG_EXIT_BT)

        await self.introduce_nodes()

        self.assertIn(self.my_peer(0), self.overlay(1).get_candidates(PEER_FLAG_EXIT_BT))
        self.assertNotIn(self.my_peer(1), self.overlay(0).get_candidates(PEER_FLAG_EXIT_BT))

    async def test_introduction_as_exit_twoway(self):
        """
        Check if two nodes can have each other as exit nodes.
        """
        self.settings(0).peer_flags.add(PEER_FLAG_EXIT_BT)
        self.settings(1).peer_flags.add(PEER_FLAG_EXIT_BT)

        await self.introduce_nodes()

        self.assertIn(self.my_peer(0), self.overlay(1).get_candidates(PEER_FLAG_EXIT_BT))
        self.assertIn(self.my_peer(1), self.overlay(0).get_candidates(PEER_FLAG_EXIT_BT))

    async def test_introduction_as_exit_noway(self):
        """
        Check if two nodes don't advertise themselves as exit node incorrectly.
        """
        await self.introduce_nodes()

        self.assertEqual(len(self.overlay(0).get_candidates(PEER_FLAG_EXIT_BT)), 0)
        self.assertEqual(len(self.overlay(1).get_candidates(PEER_FLAG_EXIT_BT)), 0)

    async def test_create_circuit(self):
        """
        Check if 1 hop circuit creation works.
        """
        self.settings(1).peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()

        # Let node 0 build tunnels of 1 hop (settings.min_circuits = settings.max_circuits = 1)
        # It should use node 1 for this
        self.overlay(0).build_tunnels(1)

        # Let the circuit creation commence
        await self.deliver_messages()

        # Node 0 should now have all of its required 1 hop circuits (1.0/100%)
        self.assertEqual(self.overlay(0).tunnels_ready(1), 1.0)
        # Node 1 has an exit socket open
        self.assertEqual(len(self.overlay(1).exit_sockets), 1)

    async def test_create_circuit_no_exit(self):
        """
        Check if 1 hop circuit creation fails without exit nodes.
        """
        await self.introduce_nodes()
        self.overlay(0).build_tunnels(1)

        # Attempt circuit creation
        await self.deliver_messages()

        # Node 0 should now have no 1 hop circuits (0.0/0%)
        self.assertEqual(self.overlay(0).tunnels_ready(1), 0.0)
        # Node 1 should not have an exit socket open
        self.assertEqual(len(self.overlay(1).exit_sockets), 0)

    async def test_create_circuit_too_many_hops(self):
        """
        Check if creating a circuit that is too long fails.
        """
        for _ in range(3):
            self.add_node_to_experiment(self.create_node())
        for node in self.nodes:
            node.overlay.settings.max_relay_early = 3

        self.overlay(1).settings.peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()
        self.overlay(0).build_tunnels(5)

        await self.deliver_messages()

        self.assertEqual(self.overlay(0).tunnels_ready(5), 0.0)

    async def test_create_circuit_relay_early_fail_hop1(self):
        """
        Check if extending a circuit using a cell with a bad relay_early flag fails at the first hop.
        """
        self.add_node_to_experiment(self.create_node())
        for node in self.nodes:
            node.overlay.settings.max_relay_early = 0

        self.overlay(1).settings.peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()

        self.overlay(0).settings.max_relay_early = 2
        self.overlay(0).build_tunnels(2)

        await self.deliver_messages()

        self.assertEqual(self.overlay(0).tunnels_ready(2), 0.0)

    async def test_create_circuit_relay_early_fail_hop2(self):
        """
        Check if extending a circuit using a cell with a bad relay_early flag fails at the second hop.
        """
        for _ in range(2):
            self.add_node_to_experiment(self.create_node())
        for node in self.nodes:
            node.overlay.settings.max_relay_early = 1

        self.overlay(1).settings.peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()

        self.overlay(0).settings.max_relay_early = 2
        self.overlay(0).build_tunnels(3)

        await self.deliver_messages()

        self.assertEqual(self.overlay(0).tunnels_ready(3), 0.0)

    async def test_create_circuit_multiple_calls(self):
        """
        Check if circuit creation is aborted when it's already building the requested circuit.
        """
        self.settings(1).peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()

        # Don't allow the exit node to answer, this keeps peer 0's circuit in EXTENDING state
        self.endpoint(1).close()
        self.overlay(0).build_tunnels(1)

        # Node 0 should have 1 circuit in the CIRCUIT_STATE_EXTENDING state
        self.assertEqual(len(self.overlay(0).find_circuits(state=CIRCUIT_STATE_EXTENDING)), 1)

        # Subsequent calls to build_circuits should not change this
        self.overlay(0).build_tunnels(1)
        self.assertEqual(len(self.overlay(0).find_circuits(state=CIRCUIT_STATE_EXTENDING)), 1)

    async def test_destroy_circuit_from_originator(self):
        """
        Check if a 2 hop circuit can be destroyed (by the exit node)
        """
        self.add_node_to_experiment(self.create_node())
        self.settings(2).peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()
        self.overlay(0).build_tunnels(2)
        await self.deliver_messages()

        # Destroy the circuit we just created using a destroy message
        await self.overlay(0).remove_circuit(list(self.overlay(0).circuits.keys())[0], destroy=1)
        await self.deliver_messages()

        self.assert_no_more_tunnels()

    async def test_destroy_circuit_from_exit(self):
        """
        Check if a 2 hop circuit can be destroyed (by the exit node)
        """
        self.add_node_to_experiment(self.create_node())
        self.settings(2).peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()
        self.overlay(0).build_tunnels(2)
        await self.deliver_messages()

        await self.overlay(2).remove_exit_socket(list(self.overlay(2).exit_sockets.keys())[0], destroy=1)
        await self.deliver_messages()

        self.assert_no_more_tunnels()

    async def test_destroy_circuit_from_relay(self):
        """
        Check if a 2 hop circuit can be destroyed (by the relay node)
        """
        self.add_node_to_experiment(self.create_node())
        self.settings(2).peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()
        self.overlay(0).build_tunnels(2)
        await self.deliver_messages()

        self.overlay(1).remove_relay(list(self.overlay(1).relay_from_to.keys())[0], destroy=1)
        await self.deliver_messages()

        self.assert_no_more_tunnels()

    async def test_destroy_circuit_bad_id(self):
        """
        Check if the correct circuit gets destroyed.
        """
        self.settings(1).peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()
        self.overlay(0).build_tunnels(1)
        await self.deliver_messages()

        # Destroy a circuit which does not exist (circuit_id + 1)
        # This should not affect other circuits
        await self.overlay(0).remove_circuit(list(self.overlay(0).circuits.keys())[0] + 1, destroy=1)
        await self.deliver_messages()

        # Node 0 should still have all of its required 1 hop circuits (1.0/100%)
        self.assertEqual(self.overlay(0).tunnels_ready(1), 1.0)
        # Node 1 still has an exit socket open
        self.assertEqual(len(self.overlay(1).exit_sockets), 1)

    async def test_tunnel_data(self):
        """
        Check if data is correctly exited.
        """
        # Listen in on communication of the target
        self.public_endpoint = UDPEndpoint()
        await self.public_endpoint.open()
        ep_listener = MockEndpointListener(self.public_endpoint)

        # Build a tunnel
        self.settings(1).peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()
        self.overlay(0).build_tunnels(1)
        await self.deliver_messages()

        # Construct a data packet
        prefix = b'\x00' * 23
        data = prefix + bytes(range(256))

        self.public_endpoint.assert_open()

        # Tunnel the data to the endpoint
        circuit = list(self.overlay(0).circuits.values())[0]
        self.overlay(0).send_data(circuit.peer, circuit.circuit_id,
                                  DomainAddress('localhost', self.public_endpoint.get_address()[1]),
                                  ('0.0.0.0', 0), data)

        future = Future()
        ep_listener.on_packet = lambda packet: ep_listener.received_packets.append(packet) or future.set_result(None)
        await future

        self.assertEqual(len(ep_listener.received_packets), 1)
        self.assertEqual(ep_listener.received_packets[0][1], data)

    async def test_two_hop_circuit(self):
        """
        Check if a two hop circuit is correctly created.

        Note that we avoid exit nodes in the relay path, so we explicitly set relay nodes to not be exits.
        """
        self.add_node_to_experiment(self.create_node())

        # Build a tunnel
        self.settings(1).peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()
        self.overlay(0).build_tunnels(2)
        await self.deliver_messages()

        self.assertEqual(self.overlay(0).tunnels_ready(2), 1.0)

    async def test_three_hop_circuit(self):
        """
        Check if a three hop circuit is correctly created.

        Note that we avoid exit nodes in the relay path, so we explicitly set relay nodes to not be exits.
        """
        self.add_node_to_experiment(self.create_node())
        self.add_node_to_experiment(self.create_node())

        # Build a tunnel
        self.settings(1).peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()
        self.overlay(0).build_tunnels(3)
        await self.deliver_messages()

        self.assertEqual(self.overlay(0).tunnels_ready(3), 1.0)

    async def test_create_two_circuit(self):
        """
        Check if multiple 1 hop circuit creation works.
        """
        self.add_node_to_experiment(self.create_node())
        self.settings(0).min_circuits = 2
        self.settings(0).max_circuits = 2
        self.settings(1).peer_flags.add(PEER_FLAG_EXIT_BT)
        self.settings(2).peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()

        # Let node 0 build tunnels of 1 hop (settings.min_circuits = settings.max_circuits = 2)
        # It should use node 1 and 2 for this
        self.overlay(0).build_tunnels(1)

        # Let the circuit creation commence
        await self.deliver_messages()

        # Node 0 should now have all of its required 1 hop circuits (1.0/100%)
        self.assertEqual(self.overlay(0).tunnels_ready(1), 1.0)
        self.assertEqual(len(self.overlay(0).circuits), 2)
        # Two exit sockets are open between node 1 and 2 (NOT evenly spread)
        self.assertEqual(len(self.overlay(1).exit_sockets) + len(self.overlay(2).exit_sockets), 2)

    async def test_reuse_partial_circuit(self):
        """
        Check if we can change the unverified hop of a circuit.
        """
        self.add_node_to_experiment(self.create_node())

        self.settings(2).peer_flags.add(PEER_FLAG_EXIT_BT)
        self.overlay(2).should_join_circuit = lambda *args: succeed(False)
        await self.introduce_nodes()
        self.overlay(0).build_tunnels(2)
        await self.deliver_messages()

        # We wanted to create circuit 0 -> 1 -> 2, but node 2 is not responding
        circuit = list(self.overlay(0).circuits.values())[0]
        self.assertEqual([h.mid for h in circuit.hops], [self.mid(1)])
        self.assertEqual(circuit.unverified_hop.mid, self.mid(2))

        # Let's add a new exit node, and retry to extend the circuit
        self.add_node_to_experiment(self.create_node())
        self.settings(3).peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()
        # Let's pretend that node 1 selected node 3 as a possible node for circuit extension
        cache = self.overlay(1).request_cache.get(u"created", circuit.circuit_id)
        cache.candidates[self.key_bin(3)] = self.my_peer(3)

        # Retry to extend the circuit
        circuit.required_exit = None
        self.overlay(0).send_extend(circuit, [self.key_bin(3)], 1)
        await self.deliver_messages()

        # Circuit should now be 0 -> 1 -> 3
        self.assertEqual([h.mid for h in circuit.hops], [self.mid(1), self.mid(3)])
        self.assertEqual(circuit.unverified_hop, None)
        self.assertEqual(self.overlay(0).tunnels_ready(2), 1.0)

    async def test_reuse_partial_circuit_first_hop(self):
        """
        Check if we can change the first unverified hop of a circuit.
        """
        self.add_node_to_experiment(self.create_node())

        self.settings(2).peer_flags.add(PEER_FLAG_EXIT_BT)
        self.overlay(1).should_join_circuit = lambda *args: succeed(False)
        await self.introduce_nodes()
        self.overlay(0).build_tunnels(2)
        await self.deliver_messages()

        # Adding the first hop fails, since hop 1 is not responding
        circuit = list(self.overlay(0).circuits.values())[0]
        self.assertEqual(circuit.hops, ())
        self.assertEqual(circuit.unverified_hop.mid, self.mid(1))

        # Let's add a new node, and retry to extend the circuit
        self.add_node_to_experiment(self.create_node())
        await self.introduce_nodes()

        # Retry to extend the circuit
        self.overlay(0).send_initial_create(circuit, [self.my_peer(3)], 1)
        await self.deliver_messages()

        # Circuit should now be 0 -> 2 -> 3
        self.assertEqual([h.mid for h in circuit.hops], [self.mid(3), self.mid(2)])
        self.assertEqual(circuit.unverified_hop, None)
        self.assertEqual(self.overlay(0).tunnels_ready(2), 1.0)

    async def test_tunnel_endpoint_anon(self):
        """
        Check if the tunnel endpoint is routing traffic correctly with anonymity enabled.
        """
        self.add_node_to_experiment(self.create_node())
        self.settings(2).peer_flags.add(PEER_FLAG_EXIT_IPV8)
        await self.introduce_nodes()
        self.overlay(0).create_circuit(1, exit_flags=[PEER_FLAG_EXIT_IPV8])
        await self.deliver_messages()

        exit_socket = list(self.overlay(2).exit_sockets.values())[0]
        self.overlay(2).exit_sockets[exit_socket.circuit_id] = MockTunnelExitSocket(exit_socket)

        send_data = self.overlay(0).send_data
        self.overlay(0).send_data = Mock(wraps=send_data)

        prefix = b'\x00\x01' + b'\x00' * 20
        self.overlay(0).endpoint = endpoint = TunnelEndpoint(self.endpoint(0))
        endpoint.set_tunnel_community(self.overlay(0))
        endpoint.set_anonymity(prefix, True)

        ep_listener = MockEndpointListener(self.endpoint(1))
        endpoint.send(self.overlay(1).my_estimated_wan, prefix + b'DATA')
        await self.deliver_messages()
        self.assertEqual(len(ep_listener.received_packets), 1)
        self.assertEqual(ep_listener.received_packets[0][1], prefix + b'DATA')
        self.overlay(0).send_data.assert_called_once()

        # When a circuit closes, sending data should fail
        self.overlay(0).send_data = Mock(wraps=send_data)
        circuit = self.overlay(0).find_circuits(exit_flags=[PEER_FLAG_EXIT_IPV8])[0]
        await self.overlay(0).remove_circuit(circuit.circuit_id)
        endpoint.send(self.overlay(1).my_estimated_wan, prefix + b'DATA')
        await self.deliver_messages()
        self.assertEqual(len(ep_listener.received_packets), 1)
        self.overlay(0).send_data.assert_not_called()

    async def test_tunnel_endpoint_no_anon(self):
        """
        Check if the tunnel endpoint is routing traffic correctly with anonymity disabled.
        """

        prefix = b'\x00' * 22
        self.overlay(0).endpoint = endpoint = TunnelEndpoint(self.endpoint(0))
        endpoint.set_tunnel_community(self.overlay(0))
        endpoint.set_anonymity(prefix, False)

        ep_listener = MockEndpointListener(self.endpoint(1))
        endpoint.send(self.overlay(1).my_estimated_wan, prefix + b'DATA')
        await self.deliver_messages()

        self.assertEqual(len(ep_listener.received_packets), 1)
        self.assertEqual(ep_listener.received_packets[0][1], prefix + b'DATA')

    async def test_tunnel_unicode_destination(self):
        """
        Check if the encoding/decoding a unicode hostname works.
        """
        self.settings(1).peer_flags.add(PEER_FLAG_EXIT_BT)
        await self.introduce_nodes()
        circuit = self.overlay(0).create_circuit(1)
        await circuit.ready

        exit_socket = list(self.overlay(1).exit_sockets.values())[0]
        mock_exit = self.overlay(1).exit_sockets[exit_socket.circuit_id] = MockTunnelExitSocket(exit_socket)
        mock_exit.sendto = Mock()

        unicode_destination = DomainAddress('JP納豆.例.jp', 1234)
        self.overlay(0).send_data(circuit.peer, circuit.circuit_id, unicode_destination, ('0.0.0.0', 0), b'')
        await self.deliver_messages()

        mock_exit.sendto.assert_called_with(b'', unicode_destination)

    async def test_test_request(self):
        """
        Check if sending test-request messages works as expected.
        """
        self.add_node_to_experiment(self.create_node())
        self.settings(1).peer_flags.add(PEER_FLAG_SPEED_TEST)
        await self.introduce_nodes()
        circuit = self.overlay(0).create_circuit(2, exit_flags=[PEER_FLAG_SPEED_TEST])
        await circuit.ready

        send_cell = self.overlay(0).send_cell
        self.overlay(0).send_cell = Mock(wraps=send_cell)
        data, _ = await self.overlay(0).send_test_request(circuit, 3, 6)
        self.assertEqual(len(self.overlay(0).send_cell.call_args[0][1].data), 3)
        self.assertEqual(len(data), 6)



import typing

from ...types import IdentityAlgorithmClass


class SchemaManager(object):
    """
    Manager for schemas: specifications of attribute disclosure algorithm parameterization.
    """

    def __init__(self) -> None:
        """
        Create a new SchemaManager: no algorithms are loaded initially.
        """
        super(SchemaManager, self).__init__()

        self.formats: typing.Dict[str, typing.Dict[str, typing.Any]] = dict()
        self.algorithms: typing.Dict[str, IdentityAlgorithmClass] = dict()

    def get_algorithm_class(self, algorithm_name: str) -> IdentityAlgorithmClass:
        """
        Get the implementation belonging to a certain algorithm name.
        These are bound to either:

         - bonehexact: for exact value matches using "Evaluating 2-DNF Formulas on Ciphertexts" by Boneh et al.
         - pengbaorange: for range proofs using "An efficient range proof scheme." by K. Peng and F. Bao.
         - irmaexact: for exact value matches using the IRMA protocol.

        :param algorithm_name: the name of the algorithm.
        """
        if algorithm_name in self.algorithms:
            return self.algorithms[algorithm_name]
        # Lazy load
        if algorithm_name == "bonehexact":
            from ..wallet.bonehexact.algorithm import BonehExactAlgorithm
            algorithm: IdentityAlgorithmClass = BonehExactAlgorithm
        elif algorithm_name == "pengbaorange":
            from ..wallet.pengbaorange.algorithm import PengBaoRangeAlgorithm
            algorithm = PengBaoRangeAlgorithm
        elif algorithm_name == "irmaexact":
            from ..wallet.irmaexact.algorithm import IRMAExactAlgorithm
            algorithm = IRMAExactAlgorithm
        else:
            raise RuntimeError(f"Attempted to load unknown proof algorithm {algorithm_name}!")
        self.algorithms[algorithm_name] = algorithm
        return algorithm

    def register_schema(self, schema_name: str, algorithm_name: str, parameters: dict) -> None:
        """
        Register a schema specification.
        Each schema is defined by an algorithm and its parameterization.

        :param schema_name: the new schema name to claim.
        :param algorithm_name: the algorithm to use (see `get_algorithm_class()`).
        :param parameters: the dictionary specifying the parameters for the algorithm.
        """
        schema = {"algorithm": algorithm_name}
        schema.update(parameters)
        self.formats[schema_name] = schema

    def register_default_schemas(self) -> None:
        """
        Register all default formats to this SchemaManager.
        You will load:

            - id_metadata: 1024 bit space "exact" value match
            - id_metadata_big: 4096 bit space "exact" value match
            - id_metadata_huge: 9216 bit space "exact" value match
            - id_metadata_range_18plus: NIZKP over a commitment, showing it lies within [0, 18]
            - id_irma_nijmegen_address_1568208470: IRMA address data match, valid until 11 Sept 2019 13:27:50 GMT
            - id_irma_nijmegen_personalData_1568208470: IRMA personal data match, valid until 11 Sept 2019 13:27:50 GMT
            - id_irma_nijmegen_ageLimits_1568208470: IRMA age data match, valid until 11 Sept 2019 13:27:50 GMT
            - id_irma_nijmegen_bsn_1568208470: IRMA bsn data match, valid until 11 Sept 2019 13:27:50 GMT
        """
        from ..default_identity_formats import FORMATS
        for schema in FORMATS:
            self.register_schema(schema, FORMATS[schema]["algorithm"], FORMATS[schema])  # type:ignore

    def get_algorithm_instance(self, schema_name: str):
        """
        Get an algorithm instance from a schema name.

        :param schema_name: the schema to instantiate the algorithm from.
        """
        schema = self.formats[schema_name]
        return self.get_algorithm_class(schema["algorithm"])(schema_name, self.formats)
